import lifelines
from lifelines import CoxPHFitter
from matplotlib import pyplot as plt

from SurvivalEVAL.Evaluations.SingleTimeCalibration import (
    integrated_calibration_index,
    one_calibration,
)

# load data
data = lifelines.datasets.load_gbsg2()
# preprocessing
data.rename(columns={"cens": "event"}, inplace=True)
data["horTh"] = data["horTh"].map({"no": 0, "yes": 1})
data["menostat"] = data["menostat"].map({"Pre": 0, "Post": 1})
data["tgrade"] = data["tgrade"].map({"I": 1, "II": 2, "III": 3})
# randomly divide the data into training and validation sets
df_train = data.sample(frac=0.7, random_state=42)  # 70% for training
df_train = df_train.reset_index(drop=True)
df_test = data.drop(df_train.index)  # remaining 30% for testing
df_test = df_test.reset_index(drop=True)
x_test = df_test.drop(columns=["time", "event"]).values

cph = CoxPHFitter()
cph.fit(df_train, duration_col="time", event_col="event")

year = 1
survs_cox = cph.predict_survival_function(x_test, times=[365 * year]).T.values.flatten()
p, hl_statistics, obs_probs, exp_probs = one_calibration(
    preds=1 - survs_cox,
    event_time=df_test["time"].values,
    event_indicator=df_test["event"].values,
    target_time=365 * year,
    num_bins=10,
    binning_strategy="H",
    method="DN",
)

ici_summary, ici_fig = integrated_calibration_index(
    preds=1 - survs_cox,
    event_time=df_test["time"].values,
    event_indicator=df_test["event"].values,
    target_time=365 * year,
    draw_figure=True,
)
print(ici_summary)
if ici_fig is not None:
    plt.show()



from typing import Optional, Tuple

import numpy as np

from SurvivalEVAL.NonparametricEstimator.SingleEvent import (
    KaplanMeierArea,
    TurnbullEstimatorLifelines,
)


def concordance(
    predicted_times: np.ndarray,
    event_times: np.ndarray,
    event_indicators: np.ndarray,
    train_event_times: Optional[np.ndarray] = None,
    train_event_indicators: Optional[np.ndarray] = None,
    method: str = "Harrell",
    ties: str = "Risk",
) -> tuple[float, float, float]:
    """
    Calculate the concordance index between the predicted survival times and the true survival times.

    Parameters
    ----------
    predicted_times: np.ndarray, shape = (n_samples,)
        The predicted survival times.
    event_times: np.ndarray, shape = (n_samples,)
        The true survival times.
    event_indicators: np.ndarray, shape = (n_samples,)
        The event indicators of the true survival times.
    train_event_times: np.ndarray, shape = (n_train_samples,)
        The true survival times of the training set.
    train_event_indicators: np.ndarray, shape = (n_train_samples,)
        The event indicators of the true survival times of the training set.
    method: str, optional (default="Harrell")
        A string indicating the method for constructing the pairs of samples.
        "Harrell": the pairs are constructed by comparing the predicted survival time of each sample with the
        event time of all other samples. The pairs are only constructed between samples with comparable
        event times. For example, if i-th sample has a censor time of 10, then the pairs are constructed by
        comparing the predicted survival time of sample i with the event time of all samples with event
        time of 10 or less.
        "Margin": the pairs are constructed between all samples. A best-guess time for the censored samples
        will be calculated and used to construct the pairs.
    ties: str, optional (default="Risk")
        A string indicating the way ties should be handled.
        Options: "None" (default), "Time", "Risk", or "All"
        "None" will throw out all ties in true survival time and all ties in predict survival times (risk scores).
        "Time" includes ties in true survival time but removes ties in predict survival times (risk scores).
        "Risk" includes ties in predict survival times (risk scores) but not in true survival time.
        "All" includes all ties.
        Note the concordance calculation is given by
        (Concordant Pairs + (Number of Ties/2))/(Concordant Pairs + Discordant Pairs + Number of Ties).

    Returns
    -------
    c_index: float
        The concordance index.
    concordant_pairs: float
        The number of concordant pairs.
    total_pairs: float
        The total number of comparable pairs.
    """
    # the scikit-survival concordance function only takes risk scores to calculate.
    # So at first we should transfer the predicted time -> risk score.
    # The risk score should be higher for subjects that live shorter (i.e. lower average survival time).

    event_indicators = event_indicators.astype(bool)

    assert (
        len(predicted_times) == len(event_times) == len(event_indicators)
    ), "The lengths of the predicted times and labels must be the same."

    if method == "Harrell":
        risks = -1 * predicted_times
        partial_weights = None
        bg_event_times = None
    elif method == "Margin":
        if train_event_times is None or train_event_indicators is None:
            error = "If 'Margin' is chosen, training set information must be provided."
            raise ValueError(error)

        train_event_indicators = train_event_indicators.astype(bool)

        km_model = KaplanMeierArea(train_event_times, train_event_indicators)
        km_linear_zero = -1 / (
            (1 - min(km_model.survival_probabilities))
            / (0 - max(km_model.survival_times))
        )
        if np.isinf(km_linear_zero):
            km_linear_zero = max(km_model.survival_times)
        predicted_times = np.clip(predicted_times, a_max=km_linear_zero, a_min=None)
        risks = -1 * predicted_times

        censor_times = event_times[~event_indicators]
        partial_weights = np.ones_like(event_indicators, dtype=float)
        partial_weights[~event_indicators] = 1 - km_model.predict(censor_times)

        best_guesses = km_model.best_guess(censor_times)
        best_guesses[censor_times > km_linear_zero] = censor_times[
            censor_times > km_linear_zero
        ]

        bg_event_times = np.copy(event_times)
        bg_event_times[~event_indicators] = best_guesses
    else:
        raise TypeError("Method for calculating concordance is unrecognized.")
    # risk_ties means predicted times are the same while true times are different.
    # time_ties means true times are the same while predicted times are different.
    # c_index, concordant_pairs, discordant_pairs, risk_ties, time_ties = metrics.concordance_index_censored(
    #     event_indicators, event_times, estimate=risk)
    c_index, concordant_pairs, discordant_pairs, risk_ties, time_ties = (
        _estimate_concordance_index(
            event_indicators,
            event_times,
            estimate=risks,
            bg_event_time=bg_event_times,
            partial_weights=partial_weights,
        )
    )
    if ties == "None":
        total_pairs = concordant_pairs + discordant_pairs
        c_index = concordant_pairs / total_pairs
    elif ties == "Time":
        total_pairs = concordant_pairs + discordant_pairs + time_ties
        concordant_pairs = concordant_pairs + 0.5 * time_ties
        c_index = concordant_pairs / total_pairs
    elif ties == "Risk":
        # This should be the same as original outputted c_index from above
        total_pairs = concordant_pairs + discordant_pairs + risk_ties
        concordant_pairs = concordant_pairs + 0.5 * risk_ties
        c_index = concordant_pairs / total_pairs
    elif ties == "All":
        total_pairs = concordant_pairs + discordant_pairs + risk_ties + time_ties
        concordant_pairs = concordant_pairs + 0.5 * (risk_ties + time_ties)
        c_index = concordant_pairs / total_pairs
    else:
        error = "Please enter one of 'None', 'Time', 'Risk', or 'All' for handling ties for concordance."
        raise TypeError(error)

    return c_index, concordant_pairs, total_pairs


def _estimate_concordance_index(
    event_indicator: np.ndarray,
    event_time: np.ndarray,
    estimate: np.ndarray,
    bg_event_time: np.ndarray = None,
    partial_weights: np.ndarray = None,
    tied_tol: float = 1e-8,
) -> tuple[float, float, float, float, float]:
    """
    Estimate the concordance index.
    This backbone of this function is borrowed from scikit-survival:
    https://github.com/sebp/scikit-survival/blob/4e664d8e4fe5e5b55006e3913f2bbabcf2455496/sksurv/metrics.py#L85-L118
    In here, we make modifications to the original function to allow for partial weights and best-guess times (margin times).

    All functions in scikit-survival are licensed under the GPLv3 License:
    https://github.com/sebp/scikit-survival/blob/master/COPYING

    Parameters
    ----------
    event_indicator: np.ndarray, shape = (n_samples,)
        The event indicators of the true survival times.
    event_time: np.ndarray, shape = (n_samples,)
        The true survival times.
    estimate: np.ndarray, shape = (n_samples,)
        The estimated risk scores. A higher score should correspond to a higher risk.
    bg_event_time: np.ndarray, shape = (n_samples,), optional (default=None)
        The best-guess event times. For uncensored samples, this should be the same as the true event times.
        For censored samples, this should be the best-guess time (margin time) for the censored samples.
    partial_weights: np.ndarray, shape = (n_samples,), optional (default=None)
        The partial weights for the censored samples.
    tied_tol: float, optional (default=1e-8)
        The tolerance for considering two times as tied.

    Returns
    -------
    c_index: float
        The concordance index.
    concordant: float
        The number of concordant pairs.
    discordant: float
        The number of discordant pairs.
    tied_risk: float
        The number of tied risk scores.
    tied_time: float
        The number of tied times.
    -------
    """
    order = np.argsort(event_time, kind="stable")

    comparable, tied_time, weight = _get_comparable(event_indicator, event_time, order)

    if partial_weights is not None:
        event_indicator = np.ones_like(event_indicator)
        comparable_2, tied_time, weight = _get_comparable(
            event_indicator, bg_event_time, order, partial_weights
        )
        for ind, mask in comparable.items():
            weight[ind][mask] = 1
        comparable = comparable_2

    if len(comparable) == 0:
        raise ValueError(
            "Data has no comparable pairs, cannot estimate concordance index."
        )

    concordant = 0
    discordant = 0
    tied_risk = 0
    numerator = 0.0
    denominator = 0.0
    for ind, mask in comparable.items():
        est_i = estimate[order[ind]]
        event_i = event_indicator[order[ind]]
        # w_i = partial_weights[order[ind]] # change this
        w_i = weight[ind]
        weight_i = w_i[order[mask]]

        est = estimate[order[mask]]

        assert event_i, (
            "got censored sample at index %d, but expected uncensored" % order[ind]
        )

        ties = np.absolute(est - est_i) <= tied_tol
        # n_ties = ties.sum()
        n_ties = np.dot(weight_i, ties.T)
        # an event should have a higher score
        con = est < est_i
        # n_con = con[~ties].sum()
        con[ties] = False
        n_con = np.dot(weight_i, con.T)

        # numerator += w_i * n_con + 0.5 * w_i * n_ties
        # denominator += w_i * mask.sum()
        numerator += n_con + 0.5 * n_ties
        denominator += np.dot(w_i, mask.T)

        tied_risk += n_ties
        concordant += n_con
        # discordant += est.size - n_con - n_ties
        discordant += np.dot(w_i, mask.T) - n_con - n_ties

    c_index = numerator / denominator
    return c_index, concordant, discordant, tied_risk, tied_time


def _get_comparable(
    event_indicator: np.ndarray,
    event_time: np.ndarray,
    order: np.ndarray,
    partial_weights: np.ndarray = None,
) -> tuple[dict, int, dict]:
    """
    Given the labels of the survival outcomes, get the comparable pairs.

    This backbone of this function is borrowed from scikit-survival:
    https://github.com/sebp/scikit-survival/blob/4e664d8e4fe5e5b55006e3913f2bbabcf2455496/sksurv/metrics.py#L57-L81
    In here, we make modifications to the original function to calculates the weights for each pair.

    All functions in scikit-survival are licensed under the GPLv3 License:
    https://github.com/sebp/scikit-survival/blob/master/COPYING

    Parameters
    ----------
    event_indicator: np.ndarray, shape = (n_samples,)
        The event indicators of the true survival times.
    event_time: np.ndarray, shape = (n_samples,)
        The true survival times.
    order: np.ndarray, shape = (n_samples,)
        The indices that would sort the event times.
    partial_weights: np.ndarray, shape = (n_samples,), optional (default=None)
        The partial weights for the censored samples.

    Returns
    -------
    comparable: dict
        A dictionary where the keys are the indices of the samples with events, and the values are boolean masks
        indicating which samples are comparable to the key sample.
    tied_time: int
        The number of tied times.
    weight: dict
        A dictionary where the keys are the indices of the samples with events, and the values are the weights
        for each comparable sample.
    """

    if partial_weights is None:
        partial_weights = np.ones_like(event_indicator, dtype=float)
    n_samples = len(event_time)
    tied_time = 0
    comparable = {}
    weight = {}

    i = 0
    while i < n_samples - 1:
        time_i = event_time[order[i]]
        end = i + 1
        while end < n_samples and event_time[order[end]] == time_i:
            end += 1

        # check for tied event times
        event_at_same_time = event_indicator[order[i:end]]
        censored_at_same_time = ~event_at_same_time

        for j in range(i, end):
            if event_indicator[order[j]]:
                mask = np.zeros(n_samples, dtype=bool)
                mask[end:] = True
                # an event is comparable to censored samples at same time point
                mask[i:end] = censored_at_same_time
                comparable[j] = mask
                tied_time += censored_at_same_time.sum()
                weight[j] = partial_weights[order] * partial_weights[order[j]]
        i = end

    return comparable, tied_time, weight


def pairwise_w(S_Li, S_Ri, eps=1e-12):
    """
    Compute pair weights w_{i<j} for interval-censored pairs.

    Parameters
    ----------
    S_Li, S_Ri : (n,) arrays
        Survival probabilities at left and right endpoints of intervals.
    eps : float
        Numerical tolerance for zero denominators.

    Returns
    -------
    w : (n, n) array
        Pairwise weights w_{i<j}.
    """
    # column/row views for broadcasting
    S_Li = S_Li[:, None]  # (n,1)
    S_Ri = S_Ri[:, None]  # (n,1)

    S_Lj = S_Li.T  # (1,n)
    S_Rj = S_Ri.T  # (1,n)

    # Monotonicity shortcuts:
    # S(max(Lj, Ri)) = min(S(Lj), S(Ri)),  S(l_max) = min(S(Lj), S(Li)),  S(r_min) = max(S(Rj), S(Ri))
    S_maxLj_ri = np.minimum(S_Lj, S_Ri)  # (n,n)
    S_lmax = np.minimum(S_Lj, S_Li)  # (n,n)
    S_rmin = np.maximum(S_Rj, S_Ri)  # (n,n)

    pos = lambda x: np.clip(x, 0.0, None)

    # Base denominator and J terms
    denom = (S_Li - S_Ri) * (S_Lj - S_Rj)  # (n,n)
    J1 = (S_Li - 0.5 * S_lmax - 0.5 * S_rmin) * pos(S_lmax - S_rmin)
    J2 = (S_Li - S_Ri) * pos(S_maxLj_ri - S_Rj)
    J = J1 + J2

    # Masks for the three edge cases:
    # A: (S_Li - S_Ri) == 0  (row-wise condition broadcast over columns)
    # B: (S_Lj - S_Rj) == 0  (column-wise condition broadcast over rows)
    A = np.isclose(S_Li - S_Ri, 0.0, atol=eps)  # (n,1)
    B = np.isclose(S_Lj - S_Rj, 0.0, atol=eps)  # (1,n)

    only_i = A & (~B)  # denominator zero due to i
    only_j = (~A) & B  # denominator zero due to j
    both = A & B  # both zero
    neither = (~A) & (~B)

    w = np.zeros_like(denom, dtype=float)

    # Case 0: neither zero -> use main formula
    if np.any(neither):
        with np.errstate(divide="ignore", invalid="ignore"):
            w[neither] = (J / np.where(np.abs(denom) > eps, denom, np.inf))[neither]

    # Case 1: (S_Li - S_Ri) == 0, (S_Lj - S_Rj) != 0
    # w = ((S(max(l_j,l_i)) - S(r_j))_+) / (S(l_j) - S(r_j))
    if np.any(only_i):
        num1 = pos(S_lmax - S_Rj)  # (n,n)
        den1 = S_Lj - S_Rj  # (n,n)
        w[only_i] = (num1 / np.clip(den1, eps, None))[only_i]

    # Case 2: (S_Lj - S_Rj) == 0, (S_Li - S_Ri) != 0
    # w = ((S(l_i) - S(min(r_i,r_j)))_+) / (S(l_i) - S(r_i))
    if np.any(only_j):
        num2 = pos(S_Li - S_rmin)  # (n,n)
        den2 = S_Li - S_Ri  # (n,1) broadcast
        w[only_j] = (num2 / np.clip(den2, eps, None))[only_j]

    # Case 3: both zero -> compare S_Li vs S_Lj
    # if S_Li > S_Lj => w=1 else 0
    if np.any(both):
        w[both] = (S_Li > S_Lj)[both].astype(float)

    # Clean up: zero diagonal, clamp tiny negatives, and (optionally) cap at 1
    np.fill_diagonal(w, 0.0)
    w = np.clip(w, 0.0, 1.0)

    return w


def concordance_ic(
    eta: np.ndarray,
    left: np.ndarray,
    right: np.ndarray,
    left_train: np.ndarray,
    right_train: np.ndarray,
    method: str = "probability",
    ties: str = "skip",
    eps: float = 1e-12,
) -> tuple[float, np.ndarray, np.ndarray]:
    """
    Concordance index for interval-censored outcomes using closed-form pair weights.

    Parameters
    ----------
    eta : np.ndarray of shape (n_sample,)
        Predicted risk scores  (higher = riskier).
    left : np.ndarray of shape (n_sample,)
        Left endpoints l_i (can be -inf).
    right : np.ndarray of shape (n_sample,)
        Right endpoints r_i (can be +inf to represent right censoring).
    left_train : np.ndarray of shape (n_train_sample,)
        Left endpoints of training data for Turnbull estimator.
    right_train : np.ndarray of shape (n_train_sample,)
        Right endpoints of training data for Turnbull estimator.
    method : {"probability", "midpoint"}, default="probability"
        Method for forming pair weights:
          - "probability": use closed-form pair weights based on Turnbull estimator.
          - "midpoint": use standard right-censored C-index on midpoint imputed times.
    ties : {"skip", "half"}, default="skip"
        How to handle ties in eta:
          - "skip": pairs with eta_i == eta_j contribute 0 to the numerator.
          - "half":  ties contribute 0.5 * w_{i<j} to the numerator.
    eps : float, default=1e-12
        Numerical guard to avoid division by ~0.

    Returns
    -------
    c_index : float
        The interval-censored concordance index as in Eq. (cindex_ic).
        Returns np.nan if the total weight denominator is 0.
    num_matrix : np.ndarray of shape (n_sample, n_sample)
        per-pair contributions to numerator,
    den_matrix : np.ndarray of shape (n_sample, n_sample)
        per-pair weights (same as weights) in denominator.
    """
    eta = np.asarray(eta, dtype=float)
    l = np.asarray(left, dtype=float)
    r = np.asarray(right, dtype=float)
    l_train = np.asarray(left_train, dtype=float)
    r_train = np.asarray(right_train, dtype=float)
    n = eta.shape[0]
    if l.shape != (n,) or r.shape != (n,):
        raise ValueError("eta, left, right must all be 1-D arrays of same length.")

    # Basic sanity
    if np.any(l > r):
        raise ValueError("Found an interval with left > right in testing data.")
    if np.any(l_train > r_train):
        raise ValueError("Found an interval with left > right in training data.")

    # train Turnbull estimator on training data
    tb = TurnbullEstimatorLifelines(l_train, r_train)

    S_l = tb.predict(l)
    S_r = tb.predict(r)
    if method == "midpoint":
        pass
    elif method == "probability":
        w = pairwise_w(S_l, S_r, eps=eps)
    else:
        raise ValueError("method must be 'probability' or 'midpoint'.")

    # Concordant matrix based on eta
    gt = (eta[:, None] > eta[None, :]).astype(float)
    if ties == "half":
        gt += 0.5 * (eta[:, None] == eta[None, :]).astype(float)
        # still zero on diagonal because w_ii is 0
    elif ties == "skip":
        pass
    else:
        raise ValueError("'ties' must be 'skip' or 'half'.")

    # Numerator and denominator (sum over ordered pairs i != j)
    num = np.sum(gt * w)
    den = np.sum(w)

    c_idx = num / den if den > 0 else float("nan")

    return c_idx, gt * w, w


def impute_times_midpoint(
    left: np.ndarray,
    right: np.ndarray,
) -> Tuple[np.ndarray, np.ndarray]:
    """
    according to (left, right] intervals, construct (t, delta) via "endpoint/midpoint imputation".

    rules:
      - Interval censoring: both L and R finite => t = (L+R)/2, delta=1
      - Right censor: L finite, R non-finite => t = L, delta= 0
      - Left censor: L non-finite, R finite => t = R, delta= 1
      - Other (both non-finite) => discard

    Returns
    -------
    t : observed/imputed time
    delta : event indicator (1=event, 0=censor)
    """
    left = np.asarray(left, dtype=float)
    right = np.asarray(right, dtype=float)

    is_L_finite = np.isfinite(left)
    is_R_finite = np.isfinite(right)

    n = left.shape[0]
    t = np.empty(n, dtype=float)
    delta = np.empty(n, dtype=int)

    # interval censoring：L limited, R limited -> use midpoint as event time
    mask_interval = is_L_finite & is_R_finite
    t[mask_interval] = 0.5 * (left[mask_interval] + right[mask_interval])
    delta[mask_interval] = 1

    # left censor: L non-finite, R finite -> use R as event time
    mask_left_cens = (~is_L_finite) & is_R_finite
    t[mask_left_cens] = right[mask_left_cens]
    delta[mask_left_cens] = 1

    # right censor: L finite, R non-finite -> use L, mark as censor
    mask_right_cens = is_L_finite & (~is_R_finite)
    t[mask_right_cens] = left[mask_right_cens]
    delta[mask_right_cens] = 0

    # other: both non-finite -> discard
    mask_bad = (~is_L_finite) & (~is_R_finite)
    if mask_bad.any():
        t[mask_bad] = np.nan
        delta[mask_bad] = -1  # mark as invalid

    # remove invalid entries
    valid = ~np.isnan(t) & (delta >= 0)
    return t[valid], delta[valid]



from typing import Optional

import numpy as np
from matplotlib import pyplot as plt
from scipy.integrate import trapezoid
from scipy.stats import chisquare, kstwobign

from SurvivalEVAL.Evaluations.util import interpolated_curve
from SurvivalEVAL.NonparametricEstimator.SingleEvent import (
    KaplanMeier,
    NelsonAalen,
    TurnbullEstimatorLifelines,
)


def d_calibration(
    pred_probs: np.ndarray, event_indicators: np.ndarray, num_bins: int = 10
) -> tuple[float, float, np.ndarray]:
    """
    Calculate the D-Calibration score.
    Parameters
    ----------
    pred_probs: np.ndarray
        The predicted survival probabilities at individual's event/censor time.
    event_indicators: np.ndarray
        The event indicators.
    num_bins: int
        The number of bins to use for the D-Calibration score.

    Returns
    -------
    statistic: float
        The test statistic of the D-Calibration test.
    pvalue: float
        The p-value of the D-Calibration test.
    combine_hist: np.ndarray
        The binning histogram of the D-Calibration test.
    """
    quantile = np.linspace(1, 0, num_bins + 1)
    censor_indicators = 1 - event_indicators

    event_probs = pred_probs[event_indicators.astype(bool)]
    event_position = np.digitize(event_probs, quantile)
    event_position[event_position == 0] = 1  # class probability==1 to the first bin

    event_hist = np.zeros([num_bins])
    for i in range(len(event_position)):
        event_hist[event_position[i] - 1] += 1

    censored_probs = pred_probs[censor_indicators.astype(bool)]

    censor_hist = np.zeros([num_bins])
    if len(censored_probs) > 0:
        for i in range(len(censored_probs)):
            partial_binning = create_censor_hist(censored_probs[i], num_bins)
            censor_hist += partial_binning

    combine_hist = event_hist + censor_hist
    statistic, pvalue = chisquare(combine_hist)
    return statistic, pvalue, combine_hist


def create_censor_hist(prob: float, num_bins: int) -> np.ndarray:
    """Create the binning histogram for a right censored instance.

    The bins are defined as follows:
    [b_0, b_1), [b_1, b_2), ..., [b_{num_bins-1}, b_num_bins] (note that the last bin is closed on both side).

    For censoring instance,
    b2 will be the infimum probability of the bin that contains S(c),
    for the bin of [b2, b3) which contains S(c), probability = (S(c) - b2) / S(c)
    for the rest of the bins, [b0, b1), [b1, b2), etc., probability = 1 / (B * S(c)), where B is the number of bins.
    :param prob: float

        The predicted probability at the censored time of a censoring instance.
    :param num_bins: int
        The number of bins to use for the D-Calibration score.
    :return:
    final_binning: np.ndarray
        The "split" histogram of this censored subject.
    """
    quantile = np.linspace(1, 0, num_bins + 1)
    censor_binning = np.zeros(num_bins)
    for i in range(num_bins):
        if prob == 1:
            censor_binning += 0.1
            break
        elif quantile[i] > prob >= quantile[i + 1]:
            first_bin = (prob - quantile[i + 1]) / prob if prob != 0 else 1
            rest_bins = 1 / (num_bins * prob) if prob != 0 else 0
            censor_binning[i] += first_bin
            censor_binning[i + 1 :] += rest_bins
            break
    # assert len(censor_binning) == num_bins, f"censor binning should have size of {num_bins}"
    return censor_binning


def d_cal_ic(
    pred_probs_left: np.ndarray, pred_probs_right: np.ndarray, num_bins: int = 10
) -> tuple[float, float, np.ndarray]:
    """
    Calculate the D-Calibration score for interval censored data.
    Parameters
    ----------
    pred_probs_left: np.ndarray
        The predicted survival probabilities at individual's left censor time.
    pred_probs_right: np.ndarray
        The predicted survival probabilities at individual's right censor time.
        For right-censored instances, the right event/censor time is infinity, so the predicted probability is 0.
    num_bins: int
        The number of bins to use for the D-Calibration score.
    Returns
    -------
    statistic: float
        The test statistic of the D-Calibration test.
    pvalue: float
        The p-value of the D-Calibration test.
    binning: np.ndarray
        The binning histogram of the D-Calibration test.
    """
    assert len(pred_probs_left) == len(
        pred_probs_right
    ), "The length of pred_probs_left and pred_probs_right should have same length."

    assert np.all(
        pred_probs_left >= pred_probs_right
    ), "The left survival probabilities should be greater than or equal to the right survival probabilities."

    assert (
        np.all(pred_probs_left >= 0)
        and np.all(pred_probs_right >= 0)
        and np.all(pred_probs_left <= 1)
        and np.all(pred_probs_right <= 1)
    ), "The predicted probabilities should be in the range [0, 1]."

    n = len(pred_probs_left)

    binning = np.zeros([num_bins])
    for i in range(n):
        partial_binning = create_interval_c_hist(
            pred_probs_left[i], pred_probs_right[i], num_bins
        )
        binning += partial_binning

    statistic, pvalue = chisquare(binning)
    return statistic, pvalue, binning


def ksd_calibration(
    pred_probs: np.ndarray,
    event_indicators: np.ndarray,
    return_details: bool = False,
) -> tuple[float, float] | tuple[float, dict]:
    """
    Calculate the K-S D-Calibration score.

    Parameters
    ----------
    pred_probs: np.ndarray
        The predicted survival probabilities at individual's event/censor time.
    event_indicators: np.ndarray
        The event indicators.
    return_details: bool
        Whether to return the detailed information including the empirical distribution and the figure.

    Returns
    -------
    pvalue: float
        The p-value of the K-S D-Calibration test.
    statistic: float
        The test statistic of the K-S D-Calibration test.
    details: dict
        The detailed information including the empirical distribution and the figure.
        - statistics: float
        - p_value: float
        - empirical_distribution: tuple (x_support, cdf_values)
        - figure: tuple (fig, ax)
    """
    assert len(pred_probs) == len(
        event_indicators
    ), "The length of pred_probs and event_indicators should have same length."

    n = len(pred_probs)
    km = KaplanMeier(pred_probs, event_indicators)
    x_support = km.survival_times
    cdf_values = km.cumulative_dens

    D_n = discrepancy_to_uniform(x_support, cdf_values)

    p_value = ks_pvalue(D_n, n)

    if return_details:
        fig, ax = plt.subplots()
        # plot the empirical CDF
        ax.step(
            x_support,
            cdf_values,
            where="post",
            label="Prediction",
            color="blue",
            linewidth=2,
        )
        # plot the ideal CDF
        ax.plot(
            [0, 1], [0, 1], label="Ideal", linestyle="--", color="grey", linewidth=2
        )
        ax.legend()
        ax.set_xlabel(r"Survival Probability $S(t_i \mid x)$")
        ax.set_ylabel("Cumulative Distribution Function")
        fig.tight_layout()

        details = {
            "statistics": D_n,
            "p_value": p_value,
            "empirical_distribution": (x_support, cdf_values),
            "figure": (fig, ax),
        }
        return p_value, details
    return p_value, D_n


def ksd_cal_ic(
    pred_probs_left: np.ndarray,
    pred_probs_right: np.ndarray,
    return_details: bool = False,
):
    """
    Calculate the K-S D-Calibration score for interval censored data.

    Parameters
    ----------
    pred_probs_left: np.ndarray
        The predicted survival probabilities at individual's left censor time.
    pred_probs_right: np.ndarray
        The predicted survival probabilities at individual's right censor time.
    return_details: bool
        Whether to return the detailed information including the empirical distribution and the figure.

    Returns
    -------
    pvalue: float
        The p-value of the K-S D-Calibration test.
    statistic: float
        The test statistic of the K-S D-Calibration test.
    details: dict
        The detailed information including the empirical distribution and the figure.
        - statistics: float
        - p_value: float
        - empirical_distribution: tuple (x_support, cdf_values)
        - figure: tuple (fig, ax)
    """
    assert len(pred_probs_left) == len(
        pred_probs_right
    ), "The length of pred_probs_left and pred_probs_right should have same length."

    assert np.all(
        pred_probs_left >= pred_probs_right
    ), "The left survival probabilities should be greater than or equal to the right survival probabilities."

    # Fit a Turnbull estimator on the predicted probabilities
    n = len(pred_probs_left)
    tb = TurnbullEstimatorLifelines(pred_probs_left, pred_probs_right)
    x_support = tb.survival_times
    cdf_values = tb.cumulative_dens

    D_n = discrepancy_to_uniform(x_support, cdf_values)

    p_value = ks_pvalue(D_n, n)

    if return_details:
        fig, ax = plt.subplots()
        # plot the empirical CDF
        ax.step(
            x_support,
            cdf_values,
            where="post",
            label="Prediction",
            color="blue",
            linewidth=2,
        )
        # plot the ideal CDF
        ax.plot(
            [0, 1], [0, 1], label="Ideal", linestyle="--", color="grey", linewidth=2
        )
        ax.legend()
        ax.set_xlabel(r"Survival Probability $S(t_i \mid x)$")
        ax.set_ylabel("Cumulative Distribution Function")
        fig.tight_layout()

        details = {
            "statistics": D_n,
            "p_value": p_value,
            "empirical_distribution": (x_support, cdf_values),
            "figure": (fig, ax),
        }
        return p_value, details

    return p_value, D_n


def create_interval_c_hist(
    prob_left: float, prob_right: float, num_bins: int
) -> np.ndarray:
    """
    Create the binning histogram for an interval censored instance.
    For interval censored instance, we have two predicted probabilities: left and right.
    The left probability is the predicted survival probability at the left censored time,
    and the right probability is the predicted survival probability at the right censored time.

    The bins are defined as follows:
    [b_0, b_1), [b_1, b_2), ..., [b_{num_bins-1}, b_num_bins] (note that the last bin is closed on both side).

    (1) If the left and right probabilities are within the same bin, then the histogram will be 1 for that bin,
    and 0 for the rest of the bins.
    (2) If the left and right probabilities are in different bins, then the histogram will be split:
        - The bin (e.g., [b3, b4))that contains the left probability will have a probability of
        (prob_left - b3) / (prob_left - prob_right)
        - The bin (e.g., [b1, b2)) that contains the right probability will have a probability of
        (b2 - prob_right) / (prob_left - prob_right)
        - The intermediate bins (e.g., [b2, b3)) will have a probability of
        1 / (num_bins * (prob_left - prob_right))

    :param prob_left: float
        The predicted probability at the left censored time of an interval censoring instance.
    :param prob_right: float
        The predicted probability at the right censored time of an interval censoring instance.
    :param num_bins: int
        The number of bins to use for the D-Calibration score.
    :return:
    hist: np.ndarray
        The "split" histogram of this interval censored subject.
    """
    # make sure the left and right probabilities are in the range [0, 1]
    if prob_right == 0:
        # if the right probability is 0, then it is a right-censored instance,
        return create_censor_hist(prob_left, num_bins)
    else:
        if prob_left < prob_right:
            # enforce the natural order for survival probs
            prob_left, prob_right = prob_right, prob_left

        edges = np.linspace(1.0, 0.0, num_bins + 1)  # descending: 1, 1-1/K, ..., 0

        left_idx = np.digitize(prob_left, edges) - 1 if prob_left < 1.0 else 0
        right_idx = np.digitize(prob_right, edges) - 1
        hist = np.zeros(num_bins, dtype=float)
        if left_idx == right_idx:
            hist[left_idx] = 1.0
            return hist
        else:
            # if the left and right probabilities are in different bins
            first_hist = (prob_left - edges[left_idx + 1]) / (prob_left - prob_right)
            hist[left_idx] += first_hist

            last_hist = (edges[right_idx] - prob_right) / (prob_left - prob_right)
            hist[right_idx] += last_hist

            # fill the intermediate bins with equal probability
            if right_idx > left_idx + 1:
                intermediate_hist = (1.0 - first_hist - last_hist) / (
                    right_idx - left_idx - 1
                )
                hist[left_idx + 1 : right_idx] = intermediate_hist

            return hist


_residual_names = {
    "CoxSnell": "Cox-Snell Residuals",
    "Modified CoxSnell-v1": "Cox-Snell Residuals",
    "Modified CoxSnell-v2": "Cox-Snell Residuals",
    "Martingale": "Martingale Residuals",
    "Deviance": "Deviance Residuals",
}


def residuals(
    pred_probs: np.ndarray,
    event_indicators: np.ndarray,
    method: str = "CoxSnell",
    draw_figure: bool = False,
) -> np.ndarray:
    """
    Calculate the residuals based on the predicted probabilities and event times.
    For details of the residuals, please refer to the Chapter 4 of the book:
    "Modelling survival data in medical research" By David Collett

    Parameters
    ----------
    pred_probs: np.ndarray
        The predicted survival probabilities at individual's event/censor time.
    event_indicators: np.ndarray
        The event indicators.
    method: str
        The method to calculate residuals. Options are "CoxSnell", "Modified CoxSnell-v1", "Modified CoxSnell-v2",
        "Martingale", "Deviance".
    draw_figure: bool
        Whether to visualize the Cox-Snell residuals and draw the cumulative hazard plot.
    Returns
    -------
    np.ndarray
        The calculated residuals.
    """
    cox_residuals = -np.log(pred_probs)

    if method == "CoxSnell":
        residuals = cox_residuals
    elif method == "Modified CoxSnell-v1" or method == "Modified CoxSnell-v2":
        # Compare with standard CoxSnell residuals, this method adds an 'excess residual' for censored instances.
        # The excess residual should also follow a unit exponential distribution, based on the lack of memory property.
        # There are two choices of excess residuals:
        # (1) use the mean of the unit exponential distribution, which is 1,
        # or (2) use the median of the unit exponential distribution, which is ln(2).
        excess_residual = 1 if method == "Modified CoxSnell-v1" else np.log(2)
        residuals = cox_residuals + excess_residual * (1 - event_indicators)
    elif method == "Martingale":
        residuals = event_indicators - cox_residuals
    elif method == "Deviance":

        def safe_log(x):
            return np.log(x + 1e-8)

        martingale_res = event_indicators - cox_residuals
        # get the sign of the martingale residuals
        sign = np.sign(martingale_res)
        residuals = sign * np.sqrt(
            -2 * (martingale_res + event_indicators * safe_log(cox_residuals))
        )
    else:
        raise ValueError("Unknown method {}".format(method))

    if draw_figure:
        cum_haz_empirical = NelsonAalen(cox_residuals, event_indicators)
        max_res = np.max(cox_residuals)
        fig, ax = plt.subplots(nrows=1, ncols=2, tight_layout=True, dpi=400)
        ax[0].plot(
            cum_haz_empirical.survival_times,
            cum_haz_empirical.cumulative_hazard,
            label="Empirical",
        )
        ax[0].plot(
            [0, max_res], [0, max_res], label="Ideal", linestyle="--", color="red"
        )
        ax[0].legend()
        ax[0].set_xlabel("Cox-Snell Residuals")
        ax[0].set_ylabel("Cumlative hazard of residuals")

        # use solid scatter points for uncensored instances and hollow scatter points for censored instances
        idx = np.arange(len(residuals))
        event_indicators = event_indicators.astype(bool)
        ax[1].scatter(
            idx[event_indicators],
            residuals[event_indicators],
            alpha=0.5,
            color="red",
            label="Event",
        )
        ax[1].scatter(
            idx[~event_indicators],
            residuals[~event_indicators],
            facecolors="none",
            edgecolors="red",
            alpha=0.5,
            label="Censored",
        )
        ax[1].set_xlabel("Index")
        ax[1].set_ylabel(_residual_names[method])
        ax[1].legend()
        plt.show()
    return residuals


def km_calibration(
    average_survival_curve: np.ndarray,
    time_coordinates: np.ndarray,
    event_times: np.ndarray,
    event_indicators: np.ndarray,
    interpolation_method: str = "Linear",
    draw_figure: bool = False,
) -> float | tuple[float, tuple[plt.Figure, plt.Axes]]:
    """
    Calculate the KM calibration score between the average prediction curve and KM curve.
    The first version of KM calibration [1] is by visual inspection of the KM curve and the average curve.
    The second version of KM calibration [2] is by calculating the KL divergence between the KM curve and the average.
    This function actually calculates the (normalized) integrated mean squared error
    between the KM curve and the average prediction curve.

    This version has three benefits over [1] and [2]:
    1. It is a single number, so it is easier to compare.
    2. This calculation is symmetric (note that KL-divergence is not).
    3. The score is between 0 and 1, where 0 means perfect calibration and 1 means worst calibration.
        And the random prediction curve will have a score of 0.25.

    Parameters
    ----------
    average_survival_curve: np.ndarray
        The average survival curves.
    time_coordinates: np.ndarray
        The time coordinates of the average survival curves.
    event_times: np.ndarray
        The event time of the test data.
    event_indicators: np.ndarray
        The event indicator of the test data.\
    draw_figure: bool
        Whether to visualize the comparison of the KM curve and average curve.

    Returns
    -------
    mse: float
        The (normalized) integrated mean squared error between the KM curve and the average prediction curve.
    fig: tuple(plt.Figure, plt.Axes)
        The matplotlib figure and axes objects for the calibration curve plot. Returned only if draw_figure
        is True.
    
    References
    ----------
    [1] Chapfuwa et al., Calibration and Uncertainty in Neural Time-to-Event Modeling， TNNLS， 2020
    [2] Yanagisawa, Proper Scoring Rules for Survival Analysis, ICML, 2023
    """
    unique_event_times = np.unique(event_times[event_indicators == 1])

    km_model = KaplanMeier(event_times, event_indicators)
    km_curve = km_model.predict(unique_event_times)

    # add 0 to the beginning of the time coordinates, and 1 to the beginning of the average_survival_curves
    unique_event_times = np.concatenate([[0], unique_event_times])
    km_curve = np.concatenate([[1], km_curve])
    if time_coordinates[0] != 0:
        time_coordinates = np.concatenate([[0], time_coordinates])
        average_survival_curve = np.concatenate([[1], average_survival_curve])

    # interpolate the average curve, so that it will have the same time coordinates as km_curve
    spline = interpolated_curve(
        time_coordinates, average_survival_curve, interpolation_method
    )
    average_survival_curve = spline(unique_event_times)
    average_survival_curve = np.clip(average_survival_curve, 0, 1)

    # integrated over the joint time coordinates
    mse = trapezoid((average_survival_curve - km_curve) ** 2, unique_event_times)
    # normalize by the maximum time coordinate
    mse /= np.max(unique_event_times)

    if draw_figure:
        fig, ax = plt.subplots(dpi=400)
        ax.plot(
            unique_event_times, average_survival_curve, label="Average Prediction Curve"
        )
        ax.plot(unique_event_times, km_curve, label="KM Curve")
        ax.fill_between(unique_event_times, average_survival_curve, km_curve, alpha=0.2)
        score_text = r"KM-Calibration$= {:.3f}$".format(mse)
        ax.plot([], [], " ", label=score_text)
        ax.legend()
        ax.set_xlabel("Time")
        ax.set_ylabel("Survival Probability")
        plt.show()
        return mse, (fig, ax)

    return mse


def coverage_ic(
    pred_l: np.ndarray,
    pred_r: np.ndarray,
    obs_l: np.ndarray,
    obs_r: np.ndarray,
    obs_l_train: Optional[np.ndarray] = None,
    obs_r_train: Optional[np.ndarray] = None,
    cov_level: float = 0.95,
    method: str = "Turnbull",
    eps: float = 1e-12,
) -> tuple[float, float, float]:
    """
    Compute the Interval-Censor Coverage (IC) metric.
    The coverage is the proportion of instances where the predicted interval [t_li, t_ui]
    contains the true event time.
    Since the true event time is interval-censored between [L_i, U_i], we consider the event time to be contained
    - 100% if the predicted interval fully covers the censoring interval [L_i, U_i]
    - 0% if there is no overlap between the predicted interval and the censoring interval [L_i, U_i]
    - partial coverage if there is a partial overlap between the predicted interval and the censoring interval [L_i, U_i]

    The partial coverage is estimated using the empirical distribution of censoring intervals from the training data.
    Like what we did for the partial weights for concordance_ic().
    That means we estimate the empirical CDF of censoring intervals [L_j, U_j] from the training data using Turnbull estimator,
    and use it to compute S(L_i), S(U_i), S(t_li), S(t_ui) for each test instance.
    The partial coverage is then calculated as:
        partial_coverage = (S(max(L_i, t_li)) - S(min(U_i, t_ui))) / (S(L_i) - S(U_i))

    Parameters
    ----------
    pred_l : np.ndarray, shape (n_samples,)
        The lower bounds of the predicted intervals.
    pred_r : np.ndarray, shape (n_samples,)
        The upper bounds of the predicted intervals.
    obs_l : np.ndarray, shape (n_samples,)
        The lower bounds of the observed censoring intervals.
    obs_r : np.ndarray, shape (n_samples,)
        The upper bounds of the observed censoring intervals.
    obs_l_train : np.ndarray, shape (n_train_samples,)
        The lower bounds of the observed censoring intervals in the training data.
    obs_r_train : np.ndarray, shape (n_train_samples,)
        The upper bounds of the observed censoring intervals in the training data.
    cov_level : float, default 0.95
        Target coverage level for calibration reference.
    method : str, default "Turnbull"
        Method to compute the coverage.
        - "Turnbull": use the empirical distribution (Turnbull estimator) of censoring intervals from the training data.
        - "linear": use linear interpolation between the left and right bounds of the censoring intervals.

    Returns
    -------
    observed_cov : float
        Average (partial) coverage across samples.
    cov_gap : float
        Difference between observed coverage and the target level (observed_cov - cov_level).
    avg_length : float
        Average length of the predicted intervals.
    """
    if pred_l.ndim != 1 or pred_r.ndim != 1:
        raise ValueError("pred_l and pred_r must be 1-dimensional arrays.")
    if obs_l.ndim != 1 or obs_r.ndim != 1:
        raise ValueError("obs_l and obs_r must be 1-dimensional arrays.")
    if not (pred_l.shape == pred_r.shape == obs_l.shape == obs_r.shape):
        raise ValueError(
            "pred_l, pred_r, obs_l, and obs_r must contain the same number of samples."
        )

    if method == "linear":
        # Linear interpolation method, assumes uniform distribution within each censoring interval
        overlap_left = np.maximum(obs_l, pred_l)
        overlap_right = np.minimum(obs_r, pred_r)

        denom = obs_r - obs_l
        numer = np.maximum(0.0, overlap_right - overlap_left)
    elif method == "Turnbull":
        # error if training is None
        if obs_l_train is None or obs_r_train is None:
            raise ValueError(
                "obs_l_train and obs_r_train must be provided for Turnbull method."
            )

        if obs_l_train.ndim != 1 or obs_r_train.ndim != 1:
            raise ValueError(
                "obs_l_train and obs_r_train must be 1-dimensional arrays."
            )

        if np.any(obs_l_train > obs_r_train):
            raise ValueError("Found training intervals with left > right.")

        tb = TurnbullEstimatorLifelines(obs_l_train, obs_r_train)

        def S(x: np.ndarray) -> np.ndarray:
            return np.asarray(tb.predict(x), dtype=float)

        S_L = S(obs_l)
        S_R = S(obs_r)

        overlap_left = np.maximum(obs_l, pred_l)
        overlap_right = np.minimum(obs_r, pred_r)

        S_overlap_left = S(overlap_left)
        S_overlap_right = S(overlap_right)

        denom = S_L - S_R
        numer = S_overlap_left - S_overlap_right
    else:
        raise ValueError("Unknown method: {}".format(method))

    coverage = np.zeros_like(denom, dtype=float)
    valid = denom > eps

    if np.any(valid):
        ratio = numer[valid] / denom[valid]
        coverage[valid] = np.clip(ratio, 0.0, 1.0)

    # Handle degenerate censoring intervals where S(L) ~= S(U)
    if np.any(~valid):
        intersects = (pred_r >= obs_l) & (pred_l <= obs_r)
        coverage[~valid] = intersects[~valid].astype(float)

    observed_cov = float(np.mean(coverage))
    cov_gap = observed_cov - float(cov_level)
    avg_length = float(np.mean(pred_r - pred_l))

    return observed_cov, cov_gap, avg_length


def discrepancy_to_uniform(
    x: np.ndarray, cdf: np.ndarray, x_support: Optional[tuple[float, float]] = None
) -> float:
    """
    Compute the Kolmogorov-Smirnov (KS) statistic for one-sample test against uniform distribution.
    The KS statistic is defined as:
        D_n = max|F_n(x) - F(x)|,
    where F_n(x) is the empirical CDF and F(x) is the CDF of the uniform distribution.
    Parameters
    ----------
    x: np.ndarray
        The support points of the empirical CDF.
    cdf: np.ndarray
        The values of the empirical CDF at the support points.
    x_support: Optional[tuple[float, float]]
        The support points of the uniform distribution. If None, it is assumed to be [0, 1].
    Returns
    -------
    D_n: float
        The KS statistic.
    """
    # sanity check
    if x.ndim != 1 or cdf.ndim != 1 or x.size != cdf.size:
        raise ValueError("x and cdf must be 1D arrays of the same length.")

    assert np.all(cdf >= 0) and np.all(
        cdf <= 1
    ), "The cdf values must be in the range [0, 1]."

    if not (np.all(np.diff(x) >= 0) and np.all(np.diff(cdf) >= 0)):
        raise ValueError("x and cdf must be nondecreasing.")

    # empirical CDF values at the support points
    F_n = cdf
    # CDF of the uniform distribution at the support points
    if x_support is None:
        F = x
    else:
        F = (x - x_support[0]) / (x_support[1] - x_support[0])

    D_plus = np.max(F_n - F)
    D_minus = np.max(F - np.concatenate(([0], F_n[:-1])))
    D_n = float(max(D_plus, D_minus))
    return D_n


def ks_pvalue(D_n: float, n: int) -> float:
    """
    Compute asymptotic KS one-sample p-value using SciPy's Kolmogorov distribution.
    This function has the same behavior as the last half part of
    scipy.stats.kstest(method='asymp') for one-sample KS test.

    But this function directly takes D_n and n as input, which is more convenient for our use case.

    The reason we do not use 'exact' method from scipy.stats.kstest is that it assumes continuous distribution,
    while in our case, the distribution estimated by empirical estimator is discrete.
    """
    lambda_n = np.sqrt(n) * D_n
    # Survival function (1 - CDF) gives the p-value directly
    return float(kstwobign.sf(lambda_n))


if __name__ == "__main__":
    ### test the KM calibration

    # # first we define the time coordinates
    # times = np.linspace(0, 100, 11)
    # # then we define the survival probabilities at each time coordinate
    # survival_probabilities = np.exp(-times / 100)
    #
    #
    # # randomly generate the event time and event indicator for 20 samples
    # num_samples = 20
    # true_t = np.random.randint(0, 100, num_samples)
    # true_e = np.random.randint(0, 2, num_samples)
    #
    # # make some random predictions between 0 and 1
    # pred_at_observed_times = np.random.rand(num_samples)
    #
    # # # calculate the KM calibration score
    # # km_calibration_score = km_calibration(survival_probabilities, times, true_t, true_e, draw_figure=True)
    # # print(km_calibration_score)
    # # calculate the Cox-Snell residuals
    # coxsnell_residuals = residuals(pred_at_observed_times, true_e, method="Martingale", draw_figure=True)

    # test interval censored D-calibration
    pred_prob_left = np.array([0.84, 0.65, 0.42, 0.75, 1, 1])
    pred_prob_right = np.array([0.72, 0.63, 0.03, 0, 0.24, 0])
    statistics, pvalue, binning = d_cal_ic(pred_prob_left, pred_prob_right, num_bins=10)



import matplotlib

matplotlib.use("Agg")

import matplotlib.pyplot as plt
import numpy as np
import pytest

from SurvivalEVAL import SurvivalEvaluator


def _sample_interval_data(rng: np.random.Generator, n_samples: int):
    for _ in range(100):
        rates = rng.uniform(0.05, 0.2, size=n_samples)
        event_times = rng.exponential(scale=1.0 / rates)
        censor_times = rng.exponential(scale=12.0, size=n_samples)
        indicators = event_times <= censor_times
        if indicators.any() and (~indicators).any():
            observed = np.minimum(event_times, censor_times)
            return rates, observed, indicators
    raise RuntimeError(
        "Failed to generate interval data with both events and censoring."
    )


@pytest.fixture
def evaluator_data():
    rng = np.random.default_rng(42)
    n_test = 80
    n_train = 120
    time_grid = np.linspace(0.0, 25.0, 80)

    train_rates, train_times, train_indicators = _sample_interval_data(rng, n_train)
    test_rates, test_times, test_indicators = _sample_interval_data(rng, n_test)

    pred_rates = test_rates * rng.uniform(0.8, 1.2, size=n_test)
    surv_curves = np.exp(-np.outer(pred_rates, time_grid))
    surv_curves = np.clip(surv_curves, 1e-8, 1.0)

    evaluator = SurvivalEvaluator(
        pred_survs=surv_curves,
        time_coordinates=time_grid,
        event_times=test_times,
        event_indicators=test_indicators,
        train_event_times=train_times,
        train_event_indicators=train_indicators,
        predict_time_method="Median",
    )

    return {
        "evaluator": evaluator,
        "time_grid": time_grid,
        "n_test": n_test,
        "train_times": train_times,
        "train_indicators": train_indicators,
        "test_times": test_times,
        "test_indicators": test_indicators,
    }


def test_prediction_utilities(evaluator_data):
    evaluator = evaluator_data["evaluator"]
    n_test = evaluator_data["n_test"]

    single_prob = evaluator.predict_probability_from_curve(8.0)
    assert single_prob.shape == (n_test,)

    per_sample_times = np.linspace(4.0, 12.0, n_test)
    per_sample_prob = evaluator.predict_probability_from_curve(per_sample_times)
    assert per_sample_prob.shape == (n_test,)

    target_grid = np.array([2.0, 6.0, 10.0])
    multi_probs = evaluator.predict_multi_probabilities_from_curve(target_grid)
    assert multi_probs.shape == (n_test, target_grid.size)

    intervals = evaluator.predict_interval(cov_level=0.8)
    assert intervals.shape == (n_test, 2)
    assert np.all(intervals[:, 0] <= intervals[:, 1])


def test_concordance_variants(evaluator_data):
    evaluator = evaluator_data["evaluator"]

    c_harrell, concordant_h, total_h = evaluator.concordance(method="Harrell")
    assert 0.0 <= c_harrell <= 1.0
    assert total_h > 0
    assert concordant_h <= total_h

    c_margin, concordant_m, total_m = evaluator.concordance(method="Margin")
    assert 0.0 <= c_margin <= 1.0
    assert total_m > 0
    assert concordant_m <= total_m


def test_auc_alias(evaluator_data):
    evaluator = evaluator_data["evaluator"]
    auc = evaluator.auc(target_time=9.0)
    auroc = evaluator.auroc(target_time=9.0)
    assert np.isclose(auc, auroc)
    assert 0.0 <= auc <= 1.0


def test_brier_scores(evaluator_data):
    evaluator = evaluator_data["evaluator"]
    brier = evaluator.brier_score(target_time=8.0, IPCW_weighted=True)
    assert np.isfinite(brier)

    times = np.array([4.0, 8.0, 12.0])
    brier_mp = evaluator.brier_score_multiple_points(
        target_times=times, IPCW_weighted=False
    )
    assert brier_mp.shape == times.shape
    assert np.all(np.isfinite(brier_mp))


def test_integrated_brier_score(evaluator_data):
    evaluator = evaluator_data["evaluator"]

    ibs_ipcw = evaluator.integrated_brier_score(num_points=10, IPCW_weighted=True)
    assert np.isfinite(ibs_ipcw)

    explicit_grid = np.linspace(2.0, 14.0, 6)
    ibs_naive = evaluator.integrated_brier_score(
        target_times=explicit_grid, IPCW_weighted=False
    )
    assert np.isfinite(ibs_naive)

    with pytest.raises(ValueError):
        evaluator.integrated_brier_score(
            num_points=5, target_times=np.array([1.0, 2.0, 3.0])
        )


def test_error_metrics(evaluator_data):
    evaluator = evaluator_data["evaluator"]

    mae_hinge = evaluator.mae(method="Hinge", weighted=False)
    mae_margin = evaluator.mae(method="Margin")
    mse_ipcwt = evaluator.mse(method="IPCW-T")
    mse_pseudo = evaluator.mse(method="Pseudo_obs")
    rmse_pseudo = evaluator.rmse(method="Pseudo_obs")

    for value in [mae_hinge, mae_margin, mse_ipcwt, mse_pseudo, rmse_pseudo]:
        assert np.isfinite(value)
        assert value >= 0

    assert np.isclose(rmse_pseudo, np.sqrt(mse_pseudo), atol=1e-6)


def test_calibration_metrics(evaluator_data):
    evaluator = evaluator_data["evaluator"]
    target_time = 8.0

    one_p, one_details = evaluator.one_calibration(
        target_time=target_time, num_bins=5, return_details=True
    )
    assert np.isfinite(one_p)
    assert "observed_probabilities" in one_details
    hist_fig, _ = one_details["histogram_plot"]
    pp_fig, _ = one_details["pp_plot"]
    plt.close(hist_fig)
    plt.close(pp_fig)

    ici_summary = evaluator.integrated_calibration_index(
        target_time=target_time, draw_figure=False
    )
    assert "ICI" in ici_summary

    d_p, d_details = evaluator.d_calibration(num_bins=6, return_details=True)
    assert np.isfinite(d_p)
    assert "histogram" in d_details
    d_hist_fig, _ = d_details["histogram_plot"]
    d_pp_fig, _ = d_details["pp_plot"]
    plt.close(d_hist_fig)
    plt.close(d_pp_fig)

    ksd_p, ksd_details = evaluator.ksd_calibration(return_details=True)
    assert np.isfinite(ksd_p)
    assert "figure" in ksd_details
    ksd_fig, _ = ksd_details["figure"]
    plt.close(ksd_fig)


def test_residuals_and_km_calibration(evaluator_data):
    evaluator = evaluator_data["evaluator"]

    residuals = evaluator.residuals(method="CoxSnell")
    assert residuals.shape[0] == evaluator_data["n_test"]

    km_cal = evaluator.km_calibration(draw_figure=False)
    assert np.isfinite(km_cal)


def test_log_rank_and_auprc(evaluator_data):
    evaluator = evaluator_data["evaluator"]

    p_value, statistic = evaluator.log_rank()
    assert np.isfinite(p_value)
    assert np.isfinite(statistic)

    auprc_value = evaluator.auprc(n_quad=64)
    assert 0.0 <= auprc_value <= 1.0


def test_pred_survs_setter_resets_cache(evaluator_data):
    evaluator = evaluator_data["evaluator"]
    original_times = evaluator.predicted_event_times.copy()
    updated_curves = evaluator.pred_survs * 0.95 + 0.05
    evaluator.pred_survs = updated_curves
    refreshed_times = evaluator.predicted_event_times
    assert not np.allclose(refreshed_times, original_times)



import numpy as np
from scipy.interpolate import interp1d

from SurvivalEVAL.Evaluations.util import check_monotonicity


def _interp_cdf_row(F_i, t_grid, t_eval, left_fill=None):
    """
    1D interpolation for a single patient's CDF:
    - left of grid: by default use F(t_min) (or 0.0 if left_fill=0.0 is passed)
    - right of grid: use 1.0 (CDF tail)
    """
    if left_fill is None:
        left_fill = F_i[0]

    f = interp1d(
        t_grid,
        F_i,
        kind="linear",
        fill_value=(left_fill, 1.0),
        bounds_error=False,
        assume_sorted=True,
    )
    return f(t_eval)


def auprc_uncensored_grid(
    pred_cdf: np.ndarray,
    time_grid: np.ndarray,
    event_times: np.ndarray,
    n_quad: int = 256,
) -> np.ndarray:
    """
    Per-patient Survival-AUPRC for an uncensored samples:
        AUPRC(y; F) = ∫_0^1 [ F(y/t) - F(y*t) ] dt
    Returns: (N,) array of scores in [0, 1].
    """
    assert check_monotonicity(
        pred_cdf
    ), "predictions_cdf must be non-decreasing over time"
    assert check_monotonicity(time_grid), "time_grid must be non-decreasing"

    event_times = np.asarray(event_times, float)
    N = event_times.shape[0]

    # Midpoint quadrature over (0, 1]
    ts = np.linspace(0.0, 1.0, n_quad + 1)
    ts_mid = 0.5 * (ts[1:] + ts[:-1])  # (Q,)
    widths = ts[1:] - ts[:-1]  # (Q,)

    scores = np.empty(N, float)
    for i in range(N):
        yi = float(event_times[i])
        # Evaluate F at transformed times
        t_right = yi / ts_mid  # y / t
        t_left = yi * ts_mid  # y * t
        Fi_right = _interp_cdf_row(pred_cdf[i], time_grid, t_right)
        Fi_left = _interp_cdf_row(pred_cdf[i], time_grid, t_left)
        integrand = Fi_right - Fi_left
        scores[i] = float(np.sum(integrand * widths))
    return scores


def auprc_right_censored_grid(
    pred_cdf: np.ndarray,
    time_grid: np.ndarray,
    censor_times: np.ndarray,
    n_quad: int = 256,
) -> np.ndarray:
    """
    Per-patient Survival-AUPRC for RIGHT-censored samples:
        AUPRC([L,∞); F) = ∫_0^1 [ 1 - F(L*t) ] dt
        (F(t_k) - F(t_k+1))/(t_k+1 - t_k)
    Returns: (Nc,) array of scores in [0, 1].
    """
    assert check_monotonicity(
        pred_cdf
    ), "predictions_cdf must be non-decreasing over time"
    assert check_monotonicity(time_grid), "time_grid must be non-decreasing"

    Nc = censor_times.shape[0]

    ts = np.linspace(0.0, 1.0, n_quad + 1)
    ts_mid = 0.5 * (ts[1:] + ts[:-1])
    widths = ts[1:] - ts[:-1]

    scores = np.empty(Nc, float)
    for i in range(Nc):
        Fi_at = _interp_cdf_row(pred_cdf[i], time_grid, float(censor_times[i]) * ts_mid)
        integrand = 1.0 - Fi_at
        scores[i] = float(np.sum(integrand * widths))
    return scores


def auprc_right_censor(
    pred_cdf: np.ndarray,
    time_grid: np.ndarray,
    event_times: np.ndarray,
    event_indicators: np.ndarray,
    n_quad: int = 256,
    return_details: bool = False,
) -> float | tuple[float, np.ndarray]:
    """
    AUPRC formula for uncensored and right-censored scenarios

    Proposed in Section 2.3 in
    Countdown Regression: Sharp and Calibrated Survival Predictions, UAI 2019
    https://proceedings.mlr.press/v115/avati20a.html

    Parameters
    ----------
    pred_cdf: np.ndarray, (n_samples, n_time_points)
        Predicted survival probabilities for the testing samples.
    time_grid:  np.ndarray, (n_time_points,)
        The time grid corresponding to the predicted CDF values.
    event_times: np.ndarray, (n_samples,)
        Observed event or censoring times for the testing samples.
    event_indicators: np.ndarray, (n_samples,)
        Binary indicators of whether the event occurred (1) or was censored (0).
    n_quad: int
        Number of quadrature points for numerical integration.
    return_details: bool
        If True, also return per-patient scores.

    Returns
    -------
    auprc:  float
        The mean AUPRC over all samples.
    scores: np.ndarray, (n_samples,)
        Per-patient AUPRC scores.
    """
    assert check_monotonicity(
        pred_cdf
    ), "predictions_cdf must be non-decreasing over time"
    assert check_monotonicity(time_grid), "time_grid must be non-decreasing"

    event_times = np.asarray(event_times, float)
    event_indicators = np.asarray(event_indicators, bool)

    # Split rows
    idx_event, idx_cens = np.where(event_indicators)[0], np.where(~event_indicators)[0]

    scores = np.empty(pred_cdf.shape[0], float)
    if idx_event.size:  # for uncensor
        scores[idx_event] = auprc_uncensored_grid(
            pred_cdf[idx_event], time_grid, event_times[idx_event], n_quad=n_quad
        )
    if idx_cens.size:  # for right censor
        scores[idx_cens] = auprc_right_censored_grid(
            pred_cdf[idx_cens], time_grid, event_times[idx_cens], n_quad=n_quad
        )

    auprc = scores.mean()
    if return_details:
        return auprc, scores
    return auprc


def auprc_ic(
    pred_cdf: np.ndarray,
    time_grid: np.ndarray,
    left: np.ndarray,
    right: np.ndarray,
    *,
    n_quad: int = 256,
    left_extrapolation_value: float = None,
    return_details: bool = False,
) -> float | tuple[float, np.ndarray]:
    """
    Per-patient Survival-AUPRC for INTERVAL-censored samples:
        AUPRC([left,right]; F) = ∫_0^1 [ F(right/t) - F(left*t) ] dt

    Proposed in Section 2.3 in
    Countdown Regression: Sharp and Calibrated Survival Predictions, UAI 2019
    https://proceedings.mlr.press/v115/avati20a.html

    Handles special cases automatically:
        - exact: left==right → reduces to uncensored AUPRC
        - right-censored: right==+inf  → term F(right/t)=1
        - left-censored:  left==0     → term F(left*t)=F(0)≈0

    Parameters
    ----------
    pred_cdf: np.ndarray, (n_samples, n_time_points)
        Predicted survival probabilities for the testing samples.
    time_grid:  np.ndarray, (n_time_points,)
        The time grid corresponding to the predicted CDF values.
    left: np.ndarray, (n_samples,)
        Left interval bounds
    right: np.ndarray, (n_samples,)
        Right interval bounds; use np.inf for right-censoring.
    n_quad: int
        Number of quadrature points for numerical integration.
    left_extrapolation_value: float, optional
        Value to use for F(0) when left=0; by default uses F(t_min).
    return_details: bool
        If True, also return per-patient scores.

    Returns:
    -------
    auprc: float
        The mean AUPRC over all samples.
    scores: np.ndarray, (n_samples,), optional
        Per-patient AUPRC scores, returned if return_details=True.
    """
    assert check_monotonicity(
        pred_cdf
    ), "predictions_cdf must be non-decreasing over time"
    assert check_monotonicity(time_grid), "time_grid must be non-decreasing"

    N = left.shape[0]

    # Midpoint quadrature over (0, 1]
    ts = np.linspace(0.0, 1.0, n_quad + 1)
    ts_mid = 0.5 * (ts[1:] + ts[:-1])  # (Q,)
    widths = ts[1:] - ts[:-1]  # (Q,)

    scores = np.empty(N, float)
    for i in range(N):
        li, ri = float(left[i]), float(right[i])

        # Right term: F(right/t).
        if np.isinf(ri):
            Fi_right = np.ones_like(
                ts_mid
            )  # If ri is +inf, this term is identically 1.
        else:
            t_right = ri / ts_mid  # can exceed grid → right=1.0
            Fi_right = _interp_cdf_row(
                pred_cdf[i], time_grid, t_right, left_fill=left_extrapolation_value
            )

        # Left term: F(L*t). If L=0 and you want F(0)=0, set left_extrapolation_value=0.0
        t_left = li * ts_mid  # can be < grid_min → left fill is used
        Fi_left = _interp_cdf_row(
            pred_cdf[i], time_grid, t_left, left_fill=left_extrapolation_value
        )

        integrand = Fi_right - Fi_left
        scores[i] = float(np.sum(integrand * widths))
    auprc = scores.mean()
    if return_details:
        return auprc, scores
    return auprc



#!/usr/bin/env python3
"""
Test cases for IBS hinge loss implementation for interval-censored survival data.
"""

import os
import sys

import numpy as np
from scipy.integrate import simpson, trapezoid

# Add the parent directory to sys.path to import SurvivalEVAL
# Insert at the beginning to prioritize local version over installed package
sys.path.insert(0, os.path.dirname(os.path.dirname(os.path.abspath(__file__))))

from SurvivalEVAL.Evaluations.BrierScore import brier_multiple_points_ic
from SurvivalEVAL.IntervalCenEvaluator import IntervalCenEvaluator


def ibs_hinge_ic(
    pred_survs: np.ndarray,  # (N, T): predicted survival probabilities
    time_coordinates: np.ndarray,  # (T,): time points corresponding to survival probabilities
    left_limits: np.ndarray,  # (N,): left limits of interval censoring
    right_limits: np.ndarray,  # (N,): right limits of interval censoring
    train_left_limits: np.ndarray = None,  # (N_train,): left limits
    train_right_limits: np.ndarray = None,  # (N_train,): right limits
    method: str = "uncensored",  # "uncensored" | "Tsouprou-marginal"
    integration_method: str = "trapezoidal",  # "trapezoidal" | "simpson"
) -> float:
    """
    Compute the Integrated Brier Score (IBS) using the hinge loss for interval-censored data.

    Parameters
    ----------
    pred_survs: np.ndarray
        Predicted survival probabilities of shape (N, T).
    time_coordinates: np.ndarray
        Time points corresponding to the survival probabilities of shape (T,).
    left_limits: np.ndarray
        Left limits of interval censoring of shape (N,).
    right_limits: np.ndarray
        Right limits of interval censoring of shape (N,).
    train_left_limits: np.ndarray, optional
        Left limits for training data of shape (N_train,).
    train_right_limits: np.ndarray, optional
        Right limits for training data of shape (N_train,).
    method: str, default: "uncensored"
        Method to handle uncertain areas. Options are "uncensored" or "Tsouprou-marginal".
    integration_method: str, default: "trapezoidal"
        Numerical integration method. Options are "trapezoidal" or "simpson".

    Returns
    -------
    float
        The computed IBS hinge loss.
    """
    brier_scores = brier_multiple_points_ic(
        pred_mat=pred_survs,
        left_limits=left_limits,
        right_limits=right_limits,
        target_times=time_coordinates,
        train_left_limits=train_left_limits,
        train_right_limits=train_right_limits,
        method=method,
    )

    if integration_method == "trapezoidal":
        ibs = trapezoid(brier_scores, time_coordinates) / (
            time_coordinates[-1] - time_coordinates[0]
        )
    elif integration_method == "simpson":
        ibs = simpson(brier_scores, time_coordinates) / (
            time_coordinates[-1] - time_coordinates[0]
        )
    else:
        raise ValueError(f"Unknown integration method: {integration_method}")

    return ibs


def test_case_1_simple_interval_censoring():
    """
    Test Case 1: Simple interval censoring with known ground truth
    """
    print("=" * 60)
    print("Test Case 1: Simple interval censoring")
    print("=" * 60)

    # Create simple test data
    n_samples = 5
    n_time_points = 10
    time_coordinates = np.linspace(0, 10, n_time_points)

    # Create predicted survival curves (decreasing probabilities over time)
    pred_survs = np.array(
        [
            [1.0, 0.9, 0.8, 0.7, 0.6, 0.5, 0.4, 0.3, 0.2, 0.1],  # Sample 1
            [1.0, 0.95, 0.85, 0.75, 0.65, 0.55, 0.45, 0.35, 0.25, 0.15],  # Sample 2
            [1.0, 0.85, 0.7, 0.55, 0.4, 0.25, 0.1, 0.05, 0.02, 0.01],  # Sample 3
            [1.0, 0.92, 0.84, 0.76, 0.68, 0.6, 0.52, 0.44, 0.36, 0.28],  # Sample 4
            [1.0, 0.88, 0.76, 0.64, 0.52, 0.4, 0.28, 0.16, 0.04, 0.01],  # Sample 5
        ]
    )

    # Interval censoring data
    left_limits = np.array([2.0, 3.0, 1.5, 4.0, 2.5])
    right_limits = np.array([4.0, 5.0, 3.0, 6.0, 4.5])

    # Training data (same structure)
    train_left_limits = np.array([1.0, 2.0, 3.0, 2.5, 3.5])
    train_right_limits = np.array([3.0, 4.0, 5.0, 4.5, 5.5])

    print(f"Time coordinates: {time_coordinates}")
    print(f"Left limits: {left_limits}")
    print(f"Right limits: {right_limits}")
    print(f"Prediction shape: {pred_survs.shape}")

    try:
        # Test IBS hinge with uncensored method
        ibs_uncensored = ibs_hinge_ic(
            pred_survs=pred_survs,
            time_coordinates=time_coordinates,
            left_limits=left_limits,
            right_limits=right_limits,
            method="uncensored",
        )
        print(f"IBS Hinge (uncensored): {ibs_uncensored:.4f}")

        # Test IBS hinge with Tsouprou-marginal method
        ibs_tsouprou = ibs_hinge_ic(
            pred_survs=pred_survs,
            time_coordinates=time_coordinates,
            left_limits=left_limits,
            right_limits=right_limits,
            train_left_limits=train_left_limits,
            train_right_limits=train_right_limits,
            method="Tsouprou-marginal",
        )
        print(f"IBS Hinge (Tsouprou-marginal): {ibs_tsouprou:.4f}")

        # Test ignore uncertain version (now integrated into main function)
        # Note: The uncensored method already ignores uncertain areas
        print(f"IBS Hinge (method already ignores uncertain): {ibs_tsouprou:.4f}")

        print("✓ Test Case 1 passed")

    except Exception as e:
        print(f"✗ Test Case 1 failed: {e}")
        import traceback

        traceback.print_exc()


def test_case_2_evaluator_integration():
    """
    Test Case 2: Test integration with IntervalCenEvaluator
    """
    print("\n" + "=" * 60)
    print("Test Case 2: IntervalCenEvaluator integration")
    print("=" * 60)

    # Create test data
    n_samples = 8
    n_time_points = 15
    time_coordinates = np.linspace(0, 15, n_time_points)

    # Create more realistic survival curves
    np.random.seed(42)
    pred_survs = np.zeros((n_samples, n_time_points))
    for i in range(n_samples):
        # Generate decreasing survival probabilities with some noise
        decay_rate = np.random.uniform(0.1, 0.3)
        pred_survs[i, :] = np.exp(-decay_rate * time_coordinates)
        # Add some noise
        noise = np.random.normal(0, 0.05, n_time_points)
        pred_survs[i, :] = np.clip(pred_survs[i, :] + noise, 0, 1)

    # Ensure monotonicity (survival probabilities should be non-increasing)
    for i in range(n_samples):
        for j in range(1, n_time_points):
            pred_survs[i, j] = min(pred_survs[i, j], pred_survs[i, j - 1])

    # Generate interval censoring data
    left_limits = np.random.uniform(2, 8, n_samples)
    right_limits = left_limits + np.random.uniform(2, 6, n_samples)

    # Training data
    train_left_limits = np.random.uniform(1, 7, n_samples)
    train_right_limits = train_left_limits + np.random.uniform(2, 5, n_samples)

    print(f"Time coordinates: {time_coordinates}")
    print(f"Left limits: {left_limits}")
    print(f"Right limits: {right_limits}")

    try:
        # Create evaluator
        evaluator = IntervalCenEvaluator(
            pred_survs=pred_survs,
            time_coordinates=time_coordinates,
            left_limits=left_limits,
            right_limits=right_limits,
            train_left_limits=train_left_limits,
            train_right_limits=train_right_limits,
        )

        # Test IBS hinge method
        ibs_result = evaluator.integrated_brier_score(method="Tsouprou-marginal")
        print(f"IBS Hinge via evaluator: {ibs_result:.4f}")

        # Test with different integration method
        ibs_simpson = evaluator.integrated_brier_score(
            method="Tsouprou-marginal", integration_method="simpson"
        )
        print(f"IBS Hinge (Simpson): {ibs_simpson:.4f}")

        # Compare with regular Brier score at a specific time
        brier_at_5 = evaluator.brier_score(target_time=5.0, method="Tsouprou-marginal")
        print(f"Brier score at t=5.0: {brier_at_5:.4f}")

        print("✓ Test Case 2 passed")

    except Exception as e:
        print(f"✗ Test Case 2 failed: {e}")
        import traceback

        traceback.print_exc()


def test_case_3_edge_cases():
    """
    Test Case 3: Edge cases and error handling
    """
    print("\n" + "=" * 60)
    print("Test Case 3: Edge cases and error handling")
    print("=" * 60)

    # Test case with point observations (no censoring)
    n_samples = 3
    n_time_points = 8
    time_coordinates = np.linspace(0, 8, n_time_points)

    pred_survs = np.array(
        [
            [1.0, 0.9, 0.8, 0.7, 0.6, 0.5, 0.4, 0.3],
            [1.0, 0.85, 0.7, 0.55, 0.4, 0.25, 0.1, 0.05],
            [1.0, 0.95, 0.9, 0.85, 0.8, 0.75, 0.7, 0.65],
        ]
    )

    # Point observations (left = right)
    event_times = np.array([3.0, 2.5, 4.5])
    left_limits = event_times
    right_limits = event_times

    train_left_limits = np.array([2.0, 3.0, 4.0])
    train_right_limits = np.array([2.0, 3.0, 4.0])

    print("Testing point observations (no censoring)...")

    try:
        evaluator = IntervalCenEvaluator(
            pred_survs=pred_survs,
            time_coordinates=time_coordinates,
            left_limits=left_limits,
            right_limits=right_limits,
            train_left_limits=train_left_limits,
            train_right_limits=train_right_limits,
        )

        # Should automatically use uncensored method
        ibs_point = evaluator.integrated_brier_score()
        print(f"IBS for point observations: {ibs_point:.4f}")

        print("✓ Point observations test passed")

    except Exception as e:
        print(f"✗ Point observations test failed: {e}")

    # Test case with very few time points
    print("\nTesting with minimal time points...")
    try:
        minimal_time = np.array([0, 5])
        minimal_pred = np.array([[1.0, 0.5], [1.0, 0.3]])
        minimal_left = np.array([2.0, 3.0])
        minimal_right = np.array([4.0, 6.0])

        ibs_minimal = ibs_hinge_ic(
            pred_survs=minimal_pred,
            time_coordinates=minimal_time,
            left_limits=minimal_left,
            right_limits=minimal_right,
            method="uncensored",
        )
        print(f"IBS with minimal time points: {ibs_minimal:.4f}")
        print("✓ Minimal time points test passed")

    except Exception as e:
        print(f"✗ Minimal time points test failed: {e}")


def test_case_4_comparison_with_paper():
    """
    Test Case 4: Simulate scenario similar to the paper
    """
    print("\n" + "=" * 60)
    print("Test Case 4: Paper-like scenario simulation")
    print("=" * 60)

    # Simulate a more realistic scenario
    np.random.seed(123)
    n_samples = 20
    n_time_points = 50
    max_time = 20
    time_coordinates = np.linspace(0, max_time, n_time_points)

    # Generate survival curves with different risk profiles
    pred_survs = np.zeros((n_samples, n_time_points))
    for i in range(n_samples):
        # High risk vs low risk patients
        if i < n_samples // 2:
            # High risk: faster decline
            hazard_rate = np.random.uniform(0.2, 0.4)
        else:
            # Low risk: slower decline
            hazard_rate = np.random.uniform(0.05, 0.15)

        pred_survs[i, :] = np.exp(-hazard_rate * time_coordinates)

    # Generate interval censoring with varying widths
    true_event_times = np.random.exponential(8, n_samples)
    interval_widths = np.random.uniform(2, 6, n_samples)

    left_limits = np.maximum(0, true_event_times - interval_widths / 2)
    right_limits = true_event_times + interval_widths / 2

    # Some right censoring (infinite right limits)
    right_censor_mask = np.random.choice([True, False], n_samples, p=[0.3, 0.7])
    right_limits[right_censor_mask] = np.inf

    # Training data
    train_true_times = np.random.exponential(7, n_samples)
    train_widths = np.random.uniform(1.5, 5, n_samples)
    train_left_limits = np.maximum(0, train_true_times - train_widths / 2)
    train_right_limits = train_true_times + train_widths / 2
    train_right_limits[np.random.choice([True, False], n_samples, p=[0.25, 0.75])] = (
        np.inf
    )

    print(f"Samples: {n_samples}, Time points: {n_time_points}")
    print(f"Max observation time: {max_time}")
    print(f"Right censoring rate: {np.mean(np.isinf(right_limits)):.2f}")
    print(
        f"Average interval width: {np.mean(right_limits[np.isfinite(right_limits)] - left_limits[np.isfinite(right_limits)]):.2f}"
    )

    try:
        # Test both IBS implementations
        ibs_standard = ibs_hinge_ic(
            pred_survs=pred_survs,
            time_coordinates=time_coordinates,
            left_limits=left_limits,
            right_limits=right_limits,
            train_left_limits=train_left_limits,
            train_right_limits=train_right_limits,
            method="Tsouprou-marginal",
        )

        # Test with different method (uncensored ignores uncertain areas)
        ibs_uncensored = ibs_hinge_ic(
            pred_survs=pred_survs,
            time_coordinates=time_coordinates,
            left_limits=left_limits,
            right_limits=right_limits,
            method="uncensored",
        )

        print(f"IBS Hinge (Tsouprou-marginal): {ibs_standard:.4f}")
        print(f"IBS Hinge (uncensored - ignores uncertain): {ibs_uncensored:.4f}")
        print(f"Difference: {abs(ibs_standard - ibs_uncensored):.4f}")

        # Test with evaluator
        evaluator = IntervalCenEvaluator(
            pred_survs=pred_survs,
            time_coordinates=time_coordinates,
            left_limits=left_limits,
            right_limits=right_limits,
            train_left_limits=train_left_limits,
            train_right_limits=train_right_limits,
        )

        ibs_evaluator = evaluator.integrated_brier_score(method="Tsouprou-marginal")
        print(f"IBS Hinge (via evaluator): {ibs_evaluator:.4f}")

        # Calculate some reference Brier scores at specific times
        brier_early = evaluator.brier_score(target_time=5, method="Tsouprou-marginal")
        brier_mid = evaluator.brier_score(target_time=10, method="Tsouprou-marginal")
        brier_late = evaluator.brier_score(target_time=15, method="Tsouprou-marginal")

        print(f"\nReference Brier scores:")
        print(f"  t=5:  {brier_early:.4f}")
        print(f"  t=10: {brier_mid:.4f}")
        print(f"  t=15: {brier_late:.4f}")

        print("✓ Test Case 4 passed")

    except Exception as e:
        print(f"✗ Test Case 4 failed: {e}")
        import traceback

        traceback.print_exc()


if __name__ == "__main__":
    print("Testing IBS Hinge Implementation for Interval-Censored Survival Data")
    print("Reference: https://arxiv.org/pdf/1806.08324")

    test_case_1_simple_interval_censoring()
    test_case_2_evaluator_integration()
    test_case_3_edge_cases()
    test_case_4_comparison_with_paper()

    print("\n" + "=" * 60)
    print("All tests completed!")
    print("=" * 60)



from typing import Optional

import numpy as np
import pandas as pd
from lifelines import WeibullAFTFitter

from SurvivalEVAL.Evaluations.custom_types import Numeric
from SurvivalEVAL.NonparametricEstimator.SingleEvent import (
    KaplanMeier,
    TurnbullEstimatorLifelines,
)


def single_brier_score(
    preds: np.ndarray,
    event_times: np.ndarray,
    event_indicators: np.ndarray,
    train_event_times: np.ndarray,
    train_event_indicators: np.ndarray,
    target_time: Optional[float] = None,
    ipcw: bool = True,
) -> float:
    """
    Calculate the Brier score at a specific time.

    Parameters
    ----------
    preds: np.ndarray, shape = (n_samples, )
        Estimated survival probabilities at the specific time for the testing samples.
    event_times: np.ndarray, shape = (n_samples, )
        Actual event/censor time for the testing samples.
    event_indicators: np.ndarray, shape = (n_samples, )
        Binary indicators of censoring for the testing samples
    train_event_times: np.ndarray, shape = (n_train_samples, )
        Actual event/censor time for the training samples.
    train_event_indicators: np.ndarray, shape = (n_train_samples, )
        Binary indicators of censoring for the training samples
    target_time: float, default: None
        The specific time point for which to estimate the Brier score.
    ipcw: bool, default: True
        Whether to use Inverse Probability of Censoring Weighting (IPCW) in the calculation.

    Returns
    -------
    brier_score: float
        Value of the brier score.
    """
    if target_time is None:
        target_time = np.median(event_times)

    event_indicators = event_indicators.astype(bool)
    # train_event_indicators = train_event_indicators.astype(bool)

    if ipcw:
        inverse_train_event_indicators = 1 - train_event_indicators
        ipc_model = KaplanMeier(train_event_times, inverse_train_event_indicators)

        ipc_pred = ipc_model.predict(event_times)
        # Catch if denominator is 0.
        ipc_pred[ipc_pred == 0] = np.inf
        # Category one calculates IPCW weight at observed time point.
        # Category one is individuals with event time lower than the time of interest and were NOT censored.
        weight_cat1 = ((event_times <= target_time) & event_indicators) / ipc_pred
        # Catch if event times goes over max training event time, i.e. predict gives NA
        weight_cat1[np.isnan(weight_cat1)] = 0
        # Category 2 is individuals whose time was greater than the time of interest (singleBrierTime)
        # contain both censored and uncensored individuals.
        weight_cat2 = (event_times > target_time) / ipc_model.predict(target_time)
        # predict returns NA if the passed-in time is greater than any of the times used to build the inverse probability
        # of censoring model.
        weight_cat2[np.isnan(weight_cat2)] = 0
    else:
        weight_cat1 = (event_times <= target_time) & event_indicators
        weight_cat2 = event_times > target_time

    b_score = (
        np.square(preds) * weight_cat1 + np.square(1 - preds) * weight_cat2
    ).mean()
    ###########################
    # Here we are ordering event times and then using predict with level.chaos = 1 which returns
    # predictions ordered by time.
    # This is from Haider's code in R, but I feel it doesn't need to be ordered by time.
    # Refer above few lines for the justified code
    ###########################
    # order_of_times = np.argsort(event_times)
    # # Catch if event times goes over max training event time, i.e. predict gives NA
    # weight_cat1 = ((event_times[order_of_times] <= target_time) & event_indicators[order_of_times]) /\
    #               ipc_model.predict(event_times[order_of_times])
    # weight_cat1[np.isnan(weight_cat1)] = 0
    # weight_cat2 = (event_times[order_of_times] > target_time) / ipc_model.predict(target_time)
    # weight_cat2[np.isnan(weight_cat2)] = 0
    #
    # survival_curves_ordered = survival_curves[order_of_times, :]
    # predict_probs = []
    # for i in range(survival_curves_ordered.shape[0]):
    #     predict_prob = predict_prob_from_curve(survival_curves_ordered[i, :], time_coordinates,
    #                                            event_times[order_of_times][i])
    #     predict_probs.append(predict_prob)
    # predict_probs = np.array(predict_probs)
    #
    # b_score = np.mean(np.square(predict_probs) * weight_cat1 + np.square(1 - predict_probs) * weight_cat2)
    return b_score


def brier_score_ic(
    preds: np.ndarray,
    left_limits: np.ndarray,
    right_limits: np.ndarray,
    train_left_limits: Optional[np.ndarray] = None,
    train_right_limits: Optional[np.ndarray] = None,
    x: Optional[np.ndarray] = None,
    x_train: Optional[np.ndarray] = None,
    target_time: Optional[Numeric] = None,
    method: str = "Tsouprou-marginal",
) -> float:
    """
    Calculate the Brier score at a specific time.

    Parameters
    ----------
    preds: np.ndarray, shape = (n_samples, )
        Estimated survival probabilities at the specific time for the testing samples.
    left_limits: np.ndarray, shape = (n_samples, )
        Actual left limit event/censor time for the testing samples.
    right_limits: np.ndarray, shape = (n_samples, )
        Actual right limit event/censor time for the testing samples.
    train_left_limits: np.ndarray, shape = (n_train_samples, )
        Actual left limit event/censor time for the training samples.
    train_right_limits: np.ndarray, shape = (n_train_samples, )
        Actual right limit event/censor time for the training samples.
    x: np.ndarray, shape = (n_samples, n_features), default: None
        Features for the testing samples. Use only when method is 'Tsouprou-conditional'.
    x_train: np.ndarray, shape = (n_train_samples, n_features), default: None
        Features for the training samples. Use only when method is 'Tsouprou-conditional'.
    target_time: numeric, default: None
        The specific time point for which to estimate the Brier score.
    method: str, default: IPCW
        Method to use for handling censoring. One of ['uncensored', 'Tsouprou-marginal', 'Tsouprou-conditional'].
        'uncensored': Treat censored data as uncensored.
        'Tsouprou-marginal': Use marginal survival probabilities based on Turnbull estimator.
        'Tsouprou-conditional': Use conditional survival probabilities based on Weibull AFT model.
    Returns
    -------
    brier_score: float
        Value of the brier score.
    """
    if target_time is None:
        tau_vals = np.concatenate(
            [left_limits, right_limits[np.isfinite(right_limits)]]
        )
        tau = np.unique(np.sort(tau_vals))
        target_time = np.median(tau)

    if method == "uncensored":
        # if the target time is within the interval, then we calculate the brier score
        # otherwise, we skip the instance
        mask = (left_limits <= target_time) & (right_limits > target_time)
        weight = 1 - mask.astype(float)
        # get the survival status at the target time
        # if the left limit is greater than the target time, then the event has not occurred, so 1
        # if the right limit is less than or equal to the target time, then the event has occurred, so 0
        survival_status = (left_limits > target_time).astype(float)
        brier_score = (np.square(preds - survival_status) * weight).sum() / weight.sum()
    elif "Tsouprou" in method:
        # method based on Sofia Tsouprou's thesis
        # Measures of discrimination and predictive accuracy for interval censored survival data
        # https://studenttheses.universiteitleiden.nl/access/item:3597164/view
        # the original method uses Weibull parametric model to estimate the survival function
        # here we give the option to use either Turnbull estimator or Weibull AFT model

        # must have training data
        if train_left_limits is None or train_right_limits is None:
            raise ValueError("Training data must be provided for Tsouprou methods.")

        if method == "Tsouprou-marginal":
            marginal_estimator = TurnbullEstimatorLifelines(
                left=train_left_limits,
                right=train_right_limits,
            )
            # get the marginal survival probabilities at the target time
            left_probs = marginal_estimator.predict(left_limits)
            right_probs = marginal_estimator.predict(right_limits)
            target_probs = marginal_estimator.predict(target_time)
        elif method == "Tsouprou-conditional":
            if x is None or x_train is None:
                raise ValueError(
                    "Features for both training and testing data must be provided for "
                    "Tsouprou-conditional method."
                )

            if x.ndim != x_train.ndim:
                raise ValueError(
                    "x and x_train must have the same number of dimensions."
                )

            train_data = {"left": train_left_limits, "right": train_right_limits}
            if x_train.ndim == 1:
                train_data["feature"] = x_train
            elif x_train.ndim == 2:
                for i in range(x_train.shape[1]):
                    train_data[f"feature_{i}"] = x_train[:, i]
            else:
                raise ValueError("x_train must be a 1-D or 2-D array.")
            train_df = pd.DataFrame(train_data)

            x_data = {}
            if x.ndim == 1:
                x_data["feature"] = x
            elif x.ndim == 2:
                for i in range(x.shape[1]):
                    x_data[f"feature_{i}"] = x[:, i]
            else:
                raise ValueError("x must be a 1-D or 2-D array.")
            x_df = pd.DataFrame(x_data)

            aft_model = WeibullAFTFitter()
            aft_model.fit_interval_censoring(train_df, "left", "right")
            # get the conditional survival probabilities at the left limit, right limit, and target time
            left_probs = aft_model.predict_survival_function(
                x_df, times=left_limits
            ).values.diagonal()
            right_probs = aft_model.predict_survival_function(
                x_df, times=right_limits
            ).values.diagonal()
            target_probs = aft_model.predict_survival_function(
                x_df, target_time
            ).values.flatten()
        else:
            raise ValueError(f"Method {method} is not supported.")
        # exam on non-bad indices
        # bad indices are those (1) the target time is within the interval and (2) the left and right survival
        # probabilities are the same, which leads to zeros in both numerator and denominator in survival_status
        bad = (
            (left_probs == right_probs)
            & (left_limits < target_time)
            & (target_time <= right_limits)
        )
        if np.sum(bad) > 0:
            left_limits = left_limits[~bad]
            right_limits = right_limits[~bad]
            preds = preds[~bad]
            left_probs = left_probs[~bad]
            right_probs = right_probs[~bad]
            if isinstance(target_time, np.ndarray):
                target_probs = target_probs[~bad]

        # supress warnings for divide by zero
        with np.errstate(divide="ignore", invalid="ignore"):
            survival_status = (target_probs - right_probs) / (left_probs - right_probs)
        survival_status[right_limits < target_time] = 0
        survival_status[left_limits >= target_time] = 1

        if np.any((survival_status < 0) | (survival_status > 1)):
            raise ValueError(
                "Calculated survival status contains values outside [0, 1]."
            )

        # calculate the brier score
        brier_score = np.mean(np.square(preds - survival_status))
    else:
        raise ValueError(f"Method {method} is not supported.")
    return brier_score


def brier_multiple_points(
    pred_mat: np.ndarray,
    event_times: np.ndarray,
    event_indicators: np.ndarray,
    train_event_times: Optional[np.ndarray],
    train_event_indicators: Optional[np.ndarray],
    target_times: np.ndarray,
    ipcw: bool = True,
) -> np.ndarray:
    """
    Calculate multiple Brier scores at multiple specific times.

    Parameters
    ----------
    pred_mat: np.ndarray, shape = (n_samples, n_time_points)
        Predicted probability array (2-D) for each instances at each time point.
    event_times: np.ndarray, shape = (n_samples, )
        Actual event/censor time for the testing samples.
    event_indicators: np.ndarray, shape = (n_samples, )
        Binary indicators of censoring for the testing samples
    train_event_times: np.ndarray, shape = (n_train_samples, )
        Actual event/censor time for the training samples.
    train_event_indicators: np.ndarray, shape = (n_train_samples, )
        Binary indicators of censoring for the training samples
    target_times: float
        The specific time points for which to estimate the Brier scores.
    ipcw: bool, default: True
        Whether to use Inverse Probability of Censoring Weighting (IPCW) in the calculation.

    Returns
    -------
    brier_scores: np.ndarray, shape = (n_time_points, )
        Values of multiple Brier scores.
    """
    if target_times.ndim != 1:
        error = "'time_grids' is not a one-dimensional array."
        raise TypeError(error)

    # bs_points_matrix = np.tile(event_times, (len(target_times), 1))
    target_times_mat = np.repeat(
        target_times.reshape(1, -1), repeats=len(event_times), axis=0
    )
    event_times_mat = np.repeat(
        event_times.reshape(-1, 1), repeats=len(target_times), axis=1
    )
    event_indicators_mat = np.repeat(
        event_indicators.reshape(-1, 1), repeats=len(target_times), axis=1
    )
    event_indicators_mat = event_indicators_mat.astype(bool)

    if ipcw:
        if train_event_times is None or train_event_indicators is None:
            raise ValueError(
                "Training event times and indicators must be provided for IPCW weighting."
            )

        inverse_train_event_indicators = 1 - train_event_indicators

        ipc_model = KaplanMeier(train_event_times, inverse_train_event_indicators)

        # Category one calculates IPCW weight at observed time point.
        # Category one is individuals with event time lower than the time of interest and were NOT censored.
        ipc_pred = ipc_model.predict(event_times_mat)
        # Catch if denominator is 0.
        ipc_pred[ipc_pred == 0] = np.inf
        weight_cat1 = (
            (event_times_mat <= target_times_mat) & event_indicators_mat
        ) / ipc_pred
        # Catch if event times goes over max training event time, i.e. predict gives NA
        weight_cat1[np.isnan(weight_cat1)] = 0
        # Category 2 is individuals whose time was greater than the time of interest (singleBrierTime)
        # contain both censored and uncensored individuals.
        ipc_target_pred = ipc_model.predict(target_times_mat)
        # Catch if denominator is 0.
        ipc_target_pred[ipc_target_pred == 0] = np.inf
        weight_cat2 = (event_times_mat > target_times_mat) / ipc_target_pred
        # predict returns NA if the passed in time is greater than any of the times used to build
        # the inverse probability of censoring model.
        weight_cat2[np.isnan(weight_cat2)] = 0
    else:
        weight_cat1 = (event_times_mat <= target_times_mat) & event_indicators_mat
        weight_cat2 = event_times_mat > target_times_mat

    ipcw_square_error_mat = (
        np.square(pred_mat) * weight_cat1 + np.square(1 - pred_mat) * weight_cat2
    )
    brier_scores = np.mean(ipcw_square_error_mat, axis=0)
    return brier_scores


def brier_multiple_points_ic(
    pred_mat: np.ndarray,
    left_limits: np.ndarray,
    right_limits: np.ndarray,
    target_times: np.ndarray,
    train_left_limits: Optional[np.ndarray] = None,
    train_right_limits: Optional[np.ndarray] = None,
    x: Optional[np.ndarray] = None,
    x_train: Optional[np.ndarray] = None,
    method: str = "Tsouprou-marginal",
) -> np.ndarray:
    """
    Compute Brier scores at multiple target times for interval-censored data.

    Parameters
    ----------
    pred_mat: np.ndarray, shape = (n_samples, n_time_points)
        Predicted survival probabilities for each sample at each target time.
        pred_mat[i, j] ~= S_hat_i(target_times[j])
    left_limits: np.ndarray, shape = (n_samples,)
        Left interval bounds L_i.
    right_limits: np.ndarray, shape = (n_samples,)
        Right interval bounds R_i. Use np.inf for right-censoring (no observed event yet).
    target_times: np.ndarray, shape = (n_time_points,)
        Time points t_j at which to evaluate the Brier score.
    train_left_limits, train_right_limits:
        Training interval bounds, required for Tsouprou-based weighting.
    x, x_train:
        Feature arrays for conditional model ('Tsouprou-conditional').
        May be 1-D (n_samples,) or 2-D (n_samples, n_features).
    method: str
        One of ['uncensored', 'Tsouprou-marginal', 'Tsouprou-conditional'].

    Returns
    -------
    brier_scores: np.ndarray, shape = (n_time_points,)
        One Brier score per target time.
    """

    # -------------------------
    # Basic shape checks
    # -------------------------
    if target_times.ndim != 1:
        raise TypeError("'target_times' must be one-dimensional.")

    n_samples = left_limits.shape[0]
    n_times = target_times.shape[0]

    if pred_mat.shape != (n_samples, n_times):
        raise ValueError(
            f"pred_mat must have shape (n_samples, n_time_points) = "
            f"({n_samples}, {n_times}), got {pred_mat.shape}"
        )

    # Broadcast helpers
    left_mat = np.repeat(
        left_limits.reshape(-1, 1), n_times, axis=1
    )  # (n_samples, n_times)
    right_mat = np.repeat(
        right_limits.reshape(-1, 1), n_times, axis=1
    )  # (n_samples, n_times)
    time_mat = np.repeat(
        target_times.reshape(1, -1), n_samples, axis=0
    )  # (n_samples, n_times)

    # ============================================================
    # Case 1: 'uncensored' (naive treating intervals like exact-ish)
    # ============================================================
    if method == "uncensored":
        # For each (i,j), define survival_status_ij in {0,1}:
        #   if t_j < L_i  -> alive -> 1
        #   if t_j >= R_i -> dead  -> 0
        #   if L_i <= t_j < R_i -> ambiguous -> exclude from averaging at that t_j
        #
        # Note: if R_i == inf (right-censored), then t_j >= R_i is False for finite t_j,
        # so status will be 1 unless t_j is in [L_i, inf) which becomes ambiguous/excluded.
        # This matches "skip samples where event time is not pinned down yet".

        ambiguous_mask = (time_mat >= left_mat) & (time_mat < right_mat)
        usable_mask = ~ambiguous_mask  # (n_samples, n_times)

        # survival_status default to 0 then overwrite alive cases:
        survival_status_mat = np.zeros_like(pred_mat, dtype=float)
        # alive if t_j < L_i
        alive_mask = time_mat < left_mat
        survival_status_mat[alive_mask] = 1.0
        # dead if t_j >= R_i and not inf
        # (already 0 there, so nothing to set)

        # squared error
        sqerr = np.square(pred_mat - survival_status_mat)

        # masked mean per time j
        # denominator per column j = number of usable samples
        usable_counts = usable_mask.sum(axis=0).astype(float)  # (n_times,)

        # avoid divide-by-zero: if no usable samples at a time point, set Brier to np.nan
        brier_scores = np.full(n_times, np.nan, dtype=float)
        for j in range(n_times):
            if usable_counts[j] > 0:
                brier_scores[j] = (sqerr[usable_mask[:, j], j]).mean()

        return brier_scores

    # ============================================================
    # Case 2/3: Tsouprou-based, which creates fractional "status"
    # ============================================================
    if "Tsouprou" in method:
        if train_left_limits is None or train_right_limits is None:
            raise ValueError("Training data must be provided for Tsouprou methods.")

        # --------------------------------------------------------
        # 2A. Fit marginal or conditional model on training data
        # --------------------------------------------------------
        if method == "Tsouprou-marginal":
            marginal_estimator = TurnbullEstimatorLifelines(
                left=train_left_limits,
                right=train_right_limits,
            )

            # Per-sample probs at L_i and R_i
            left_probs = marginal_estimator.predict(left_limits)  # (n_samples,)
            right_probs = marginal_estimator.predict(right_limits)  # (n_samples,)

            # Per-time probs S(t_j). Same for every sample.
            target_probs_vec = marginal_estimator.predict(target_times)  # (n_times,)

            # Broadcast:
            left_probs_mat = np.repeat(left_probs.reshape(-1, 1), n_times, axis=1)
            right_probs_mat = np.repeat(right_probs.reshape(-1, 1), n_times, axis=1)
            target_probs_mat = np.repeat(
                target_probs_vec.reshape(1, -1), n_samples, axis=0
            )

        elif method == "Tsouprou-conditional":
            if x is None or x_train is None:
                raise ValueError(
                    "x and x_train must be provided for Tsouprou-conditional."
                )

            # build train_df
            train_data = {"left": train_left_limits, "right": train_right_limits}
            if x_train.ndim == 1:
                train_data["feature"] = x_train
            elif x_train.ndim == 2:
                for k in range(x_train.shape[1]):
                    train_data[f"feature_{k}"] = x_train[:, k]
            else:
                raise ValueError("x_train must be a 1-D or 2-D array.")
            train_df = pd.DataFrame(train_data)

            # build x_df for test
            x_data = {}
            if x.ndim == 1:
                x_data["feature"] = x
            elif x.ndim == 2:
                for k in range(x.shape[1]):
                    x_data[f"feature_{k}"] = x[:, k]
            else:
                raise ValueError("x must be a 1-D or 2-D array.")
            x_df = pd.DataFrame(x_data)

            aft_model = WeibullAFTFitter()
            aft_model.fit_interval_censoring(train_df, "left", "right")

            # left_probs[i]   = S_i(L_i)
            # right_probs[i]  = S_i(R_i)
            # We'll grab these by diagonal extraction, same as single-time code:
            left_sf_df = aft_model.predict_survival_function(
                x_df, times=left_limits
            )  # rows=times, cols=samples
            right_sf_df = aft_model.predict_survival_function(x_df, times=right_limits)
            left_probs = left_sf_df.values.diagonal()  # (n_samples,)
            right_probs = right_sf_df.values.diagonal()  # (n_samples,)

            # target_probs_mat[i,j] = S_i(t_j)
            full_sf_df = aft_model.predict_survival_function(x_df, times=target_times)
            # full_sf_df: rows are t_j, cols are samples i
            target_probs_mat = full_sf_df.T.values  # shape (n_samples, n_times)

            # broadcast left/right
            left_probs_mat = np.repeat(left_probs.reshape(-1, 1), n_times, axis=1)
            right_probs_mat = np.repeat(right_probs.reshape(-1, 1), n_times, axis=1)

        else:
            raise ValueError(f"Method {method} is not supported.")

        # --------------------------------------------------------
        # 2B. Build Y_mat (fractional survival status) for every (i,j)
        # --------------------------------------------------------
        # We'll follow your single-time logic:
        # survival_status = (S(t) - S(R)) / (S(L) - S(R))
        # Then override based on position of t relative to [L,R].
        #
        # Edge cases:
        # - If denominator is 0 *and* t is strictly inside (L,R], we can't define status. We'll drop those cells.
        # - After override:
        #       if t_j > R_i  -> 0
        #       if t_j <= L_i -> 1
        # This guarantees values in [0,1].

        denom = left_probs_mat - right_probs_mat

        with np.errstate(divide="ignore", invalid="ignore"):
            survival_status_mat = (target_probs_mat - right_probs_mat) / denom

        # Apply boundary overrides
        after_right_mask = time_mat > right_mat
        before_left_mask = time_mat <= left_mat
        survival_status_mat[after_right_mask] = 0.0
        survival_status_mat[before_left_mask] = 1.0

        # Identify "bad" cells:
        # bad if denom == 0 AND t_j is within (L_i, R_i] (i.e. genuinely ambiguous)
        inside_mask = (time_mat > left_mat) & (time_mat <= right_mat)
        bad_mask = (denom == 0.0) & inside_mask

        # sanity check: any out-of-range due to numerical issues?
        oob_mask = (survival_status_mat < 0.0) | (survival_status_mat > 1.0)
        if np.any(oob_mask & ~bad_mask):
            raise ValueError(
                "Calculated survival status contains values outside [0,1] for some non-bad entries."
            )

        # We'll exclude bad_mask entries from the averaging for their column.
        usable_mask = ~bad_mask

        # --------------------------------------------------------
        # 2C. Compute squared error and average per time
        # --------------------------------------------------------
        sqerr = np.square(pred_mat - survival_status_mat)

        brier_scores = np.full(n_times, np.nan, dtype=float)
        usable_counts = usable_mask.sum(axis=0).astype(float)

        for j in range(n_times):
            if usable_counts[j] > 0:
                brier_scores[j] = (sqerr[usable_mask[:, j], j]).mean()

        return brier_scores

    raise ValueError(f"Method {method} is not supported.")



from typing import Dict, Literal, Optional, Tuple

import numpy as np


def right_censor_to_interval(
    event_indicators: np.ndarray,  # (N,) bool: True=event, False=right-censor
    observed_times: np.ndarray,  # (N,) float: event or censor time
    max_t=np.inf,  # max time
):
    """
    Convert (event_indicators, observed_times) to lifelines-style interval data.

    Rules:
      - Event at time t:
          left = event time t
      - Right-censored at time c:
          left = max_t
    """
    e = np.asarray(event_indicators, bool)
    y = np.asarray(observed_times, float)
    N = y.shape[0]

    left = np.empty(N, dtype=float)
    right = np.empty(N, dtype=float)

    # small tolerance for alignment checks
    eps = 1e-12

    for i in range(N):
        t = y[i]

        if e[i]:
            # event case
            left[i] = t
            right[i] = t
        else:
            # right-censored case
            left[i] = t
            right[i] = max_t

    return left, right


def _visits_fixed(end: float, step: float) -> np.ndarray:
    if end <= 0:
        return np.array([0.0])
    start = 0.0
    grid = np.arange(start, end, step)
    if grid.size == 0 or grid[-1] != end:
        grid = np.concatenate([grid, [end]])
    return np.unique(np.clip(grid, 0.0, end))


def _visits_poisson(rng: np.random.Generator, end: float, rate: float) -> np.ndarray:
    if end <= 0:
        return np.array([0.0])
    K = rng.poisson(lam=max(rate * end, 0.0))
    times = rng.uniform(0.0, end, size=K)
    times = np.concatenate(([0.0], times))
    times = np.concatenate((times, [end]))
    times.sort()
    return np.unique(times)


def _visits_lognormal(
    rng: np.random.Generator, end: float, mean: float, sigma: float
) -> np.ndarray:
    times = [0.0]
    t = float(rng.lognormal(mean, sigma))
    while t < end:
        times.append(t)
        t += float(rng.lognormal(mean, sigma))
    times.append(end)
    return np.unique(np.array(times))


def _visits_exp(rng, end, rate: float = 1.0):
    t = 0.0
    times = [0.0]
    while True:
        t += rng.exponential(1.0 / rate)
        if t >= end:
            break
        times.append(t)
    times.append(end)
    return np.array(times)


def _hawkes_times(
    rng: np.random.Generator,
    mu: float = 0.2,
    alpha: float = 0.3,
    beta: float = 1.5,
    end: float = 1000,
    max_events: int = 100,
) -> np.ndarray:
    """
    Simulate Hawkes with intensity: λ(t) = μ + α * sum_k exp(-β (t - t_k)).
    Returns sorted event times in (0, T].
    """
    assert mu >= 0 and alpha >= 0 and beta > 0 and end >= 0
    t, g = 0.0, 0.0  # g = sum exp(-β (t - t_k))
    lam = mu + alpha * g
    times = []

    while t < end and len(times) < max_events:
        if lam <= 0:
            break
        w = rng.exponential(1.0 / lam)  # candidate gap
        t_cand = t + w
        if t_cand > end:
            break
        # decay g over gap
        g *= np.exp(-beta * w)
        lam_cand = mu + alpha * g
        # accept with prob lam_cand / lam
        if rng.random() * lam <= lam_cand and lam_cand > 0:
            # event at t_cand
            times.append(t_cand)
            g += 1.0  # kernel at 0 is 1
            t = t_cand
            lam = mu + alpha * g
        else:
            t = t_cand
            lam = lam_cand

    return np.array(times, dtype=float)


def interval_censor_DGP_from_synthetic_times(
    event_times: np.ndarray,  # (N,) true event time
    censoring_times: np.ndarray,  # (N,) right-censor/admin end per subject
    *,
    method: str = "fixed",
    # fixed
    params: Optional[Dict[str, float]] = None,
    seed: int = 42,
) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    """
    Convert (event_times, censoring_times) into interval-censored bounds (left, right):
      - If event in (visit_j, visit_{j+1}]: left=visit_j, right=visit_{j+1} (interval-censor)
      - If event == a visit:             left=right=event (uncensor)
      - If event > last visit:           left=last visit, right=+inf (right-censor)
    Returns: (left, right, n_visits)
    """
    rng = np.random.default_rng(seed)
    e = np.asarray(event_times, float)
    c = np.asarray(censoring_times, float)
    assert e.shape == c.shape
    N = e.shape[0]

    left = np.empty(N, float)
    right = np.empty(N, float)
    n_visits = np.empty(N, int)

    for i in range(N):
        end_i = float(c[i])
        end_i = max(end_i, 0.0)

        if method == "fixed":
            step = params["step"] if "step" in params else 1.0
            visits = _visits_fixed(end=end_i, step=step)
        elif method == "poisson":
            rate = params["rate"] if "rate" in params else 1.0
            visits = _visits_poisson(rng, end=end_i, rate=rate)
        elif method == "lognormal":
            ln_mean = params["ln_mean"] if "ln_mean" in params else 0.0
            ln_sigma = params["ln_sigma"] if "ln_sigma" in params else 1.0
            visits = _visits_lognormal(rng, end=end_i, mean=ln_mean, sigma=ln_sigma)
        elif method == "exp":
            rate = params["rate"] if "rate" in params else 2.0
            visits = _visits_exp(rng, end=end_i, rate=rate)
        elif method == "hawkes":
            mu = params["mu"] if "mu" in params else 0.2
            alpha = params["alpha"] if "alpha" in params else 0.3
            beta = params["beta"] if "beta" in params else 1.5
            visits = _hawkes_times(rng, end=end_i, mu=mu, alpha=alpha, beta=beta)
        else:
            raise ValueError(
                "method must be 'fixed'|'poisson'|'lognormal'|'hawkes'|'exp'"
            )

        n_visits[i] = visits.size

        t = float(e[i])
        # right censor
        if t > c[i] or t > visits[-1]:
            left[i] = c[i]
            right[i] = np.inf
            continue

        # interval censor: find j let visits[j] < t <= visits[j+1]
        j = int(np.searchsorted(visits, t, side="right") - 1)
        # print (visits, j, t)
        if np.isclose(t, float(visits[j])):
            left[i] = right[i] = t  # exact
        elif np.isclose(t, float(visits[j + 1])):
            left[i] = right[i] = t  # exact
        else:
            L, U = float(visits[j]), float(visits[j + 1])
            left[i], right[i] = L, U

    return left, right, n_visits



import numpy as np
import pytest

from SurvivalEVAL.Evaluations.BrierScore import (
    brier_multiple_points,
    brier_multiple_points_ic,
    brier_score_ic,
    single_brier_score,
)


@pytest.mark.parametrize("ipcw", [True, False])
def test_brier_multiple_points_matches_single(ipcw: bool) -> None:
    pred_mat = np.array(
        [
            [0.85, 0.70, 0.55],
            [0.80, 0.60, 0.40],
            [0.65, 0.50, 0.35],
            [0.60, 0.45, 0.30],
        ]
    )
    event_times = np.array([2.0, 5.0, 9.0, 12.0])
    event_indicators = np.array([1, 0, 1, 1])
    train_event_times = np.array([1.0, 2.5, 4.5, 6.0, 8.0, 9.5, 11.0, 13.0])
    train_event_indicators = np.array([1, 1, 0, 1, 1, 0, 1, 1])
    target_times = np.array([3.0, 6.0, 10.0])

    multi_scores = brier_multiple_points(
        pred_mat=pred_mat,
        event_times=event_times,
        event_indicators=event_indicators,
        train_event_times=train_event_times,
        train_event_indicators=train_event_indicators,
        target_times=target_times,
        ipcw=ipcw,
    )

    single_scores = np.array(
        [
            single_brier_score(
                preds=pred_mat[:, idx],
                event_times=event_times,
                event_indicators=event_indicators,
                train_event_times=train_event_times,
                train_event_indicators=train_event_indicators,
                target_time=target_times[idx],
                ipcw=ipcw,
            )
            for idx in range(target_times.size)
        ]
    )

    np.testing.assert_allclose(multi_scores, single_scores, rtol=1e-6, atol=1e-8)


def test_brier_multiple_points_ic_matches_single_uncensored() -> None:
    pred_mat = np.array(
        [
            [0.90, 0.75, 0.50],
            [0.85, 0.65, 0.45],
            [0.70, 0.55, 0.35],
            [0.60, 0.40, 0.25],
        ]
    )
    left_limits = np.array([0.0, 2.0, 4.0, 6.0])
    right_limits = np.array([1.0, 3.5, np.inf, 8.5])
    target_times = np.array([0.5, 2.5, 7.0])

    multi_scores = brier_multiple_points_ic(
        pred_mat=pred_mat,
        left_limits=left_limits,
        right_limits=right_limits,
        target_times=target_times,
        method="uncensored",
    )

    single_scores = np.array(
        [
            brier_score_ic(
                preds=pred_mat[:, idx],
                left_limits=left_limits,
                right_limits=right_limits,
                target_time=target_times[idx],
                method="uncensored",
            )
            for idx in range(target_times.size)
        ]
    )

    np.testing.assert_allclose(
        multi_scores, single_scores, rtol=1e-6, atol=1e-8, equal_nan=True
    )



from typing import Optional

import numpy as np
from tqdm import trange

from SurvivalEVAL.NonparametricEstimator.SingleEvent import KaplanMeierArea, km_mean


def mean_error(
    predicted_times: np.ndarray,
    event_times: np.ndarray,
    event_indicators: np.ndarray,
    train_event_times: Optional[np.ndarray] = None,
    train_event_indicators: Optional[np.ndarray] = None,
    error_type: str = "absolute",
    method: str = "Hinge",
    weighted: bool = True,
    log_scale: bool = False,
    verbose: bool = False,
    truncated_time=None,
) -> float:
    """
    Calculate the mean absolute/squared error score for the predicted survival times.

    Parameters
    ----------
    predicted_times: np.ndarray, shape = (n_samples, )
        Predicted survival times for the testing samples
    event_times: np.ndarray, shape = (n_samples, )
        Actual event/censor time for the testing samples.
    event_indicators: np.ndarray, shape = (n_samples, )
        Binary indicators of censoring for the testing samples
    train_event_times: np.ndarray, shape = (n_train_samples, )
        Actual event/censor time for the training samples.
    train_event_indicators: np.ndarray, shape = (n_train_samples, )
        Binary indicators of censoring for the training samples
    error_type: string, default: "absolute"
        Type of mean error to use. Options are "absolute" and "squared".
    method: string, default: "Hinge"
        Method of handling censorship.
        Options are "Uncensored", "Hinge", "Margin", "IPCW-T", "IPCW-D", "Pseudo_obs", and "Pseudo_obs_pop"
    weighted: boolean, default: True
        Whether to use weighting scheme for MAE.
        If true, each best guess value / surrogate value will have a confidence weight = 1/ (1 - KM(censoring time)).
    log_scale: boolean, default: False
        Whether to use log scale for the loss function.
    verbose: boolean, default: False
        Whether to show the progress bar.
    truncated_time: float, default: None
        the truncated time.

    Returns
    -------
    Value for the calculated MAE score.
    """
    event_indicators = event_indicators.astype(bool)
    n_test = event_times.size
    if train_event_indicators is not None:
        train_event_indicators = train_event_indicators.astype(bool)

    # calculate the weighting for each sample
    if method in ["Margin", "IPCW-T", "IPCW-D", "Pseudo_obs", "Pseudo_obs_pop"]:
        if train_event_times is None or train_event_indicators is None:
            raise ValueError(
                "If method is '{}', training set values must be included.".format(
                    method
                )
            )

        km_model = KaplanMeierArea(train_event_times, train_event_indicators)
        km_linear_zero = km_model.km_linear_zero
        if np.isinf(km_linear_zero):
            km_linear_zero = max(km_model.survival_times)
        # predicted_times = np.clip(predicted_times, a_max=km_linear_zero, a_min=None)

        censor_times = event_times[~event_indicators]
        weights = np.ones(n_test)
        if weighted:
            weights[~event_indicators] = 1 - km_model.predict(censor_times)

    # set the error function
    if error_type == "absolute":
        error_func = np.abs
    elif error_type == "squared":
        error_func = np.square
    else:
        raise TypeError(
            "Please enter one of 'absolute' or 'squared' for calculating error."
        )

    if method == "Uncensored":
        # only use uncensored data
        if log_scale:
            errors = np.log(event_times[event_indicators]) - np.log(
                predicted_times[event_indicators]
            )
        else:
            errors = event_times[event_indicators] - predicted_times[event_indicators]
        return error_func(errors).mean()
    elif method == "Hinge":
        # consider only the early prediction error
        # if prediction is higher than the censored time, it is not penalized
        weights = np.ones(predicted_times.size)
        if weighted:
            if train_event_times is None or train_event_indicators is None:
                raise ValueError(
                    "If 'weighted' is True for calculating Hinge, training set values must be included."
                )
            km_model = KaplanMeierArea(train_event_times, train_event_indicators)
            censor_times = event_times[~event_indicators]
            weights[~event_indicators] = 1 - km_model.predict(censor_times)

        if log_scale:
            errors = np.log(event_times) - np.log(predicted_times)
        else:
            errors = event_times - predicted_times
        errors[~event_indicators] = np.maximum(errors[~event_indicators], 0)
        return np.average(error_func(errors), weights=weights)
    elif method == "Margin":
        # The L1-margin method proposed by https://www.jmlr.org/papers/v21/18-772.html
        # Calculate the best guess survival time given the KM curve and censoring time of that patient
        best_guesses = km_model.best_guess(censor_times)
        best_guesses[censor_times > km_linear_zero] = censor_times[
            censor_times > km_linear_zero
        ]

        if truncated_time:
            best_guesses = np.clip(best_guesses, a_max=truncated_time, a_min=None)
            predicted_times = np.clip(predicted_times, a_max=truncated_time, a_min=None)
            event_times = np.clip(event_times, a_max=truncated_time, a_min=None)

        errors = np.empty(predicted_times.size)
        if log_scale:
            errors[event_indicators] = np.log(event_times[event_indicators]) - np.log(
                predicted_times[event_indicators]
            )
            errors[~event_indicators] = np.log(best_guesses) - np.log(
                predicted_times[~event_indicators]
            )
        else:
            errors[event_indicators] = (
                event_times[event_indicators] - predicted_times[event_indicators]
            )
            errors[~event_indicators] = (
                best_guesses - predicted_times[~event_indicators]
            )
        return np.average(error_func(errors), weights=weights)
    elif method == "IPCW-T":
        # This is the IPCW-T method from https://arxiv.org/pdf/2306.01196.pdf
        # Calculate the best guess time (surrogate time) based on the subsequent uncensored subjects
        best_guesses = np.empty(shape=n_test)
        for i in range(n_test):
            if event_indicators[i] == 1:
                best_guesses[i] = event_times[i]
            else:
                # Numpy will throw a warning if afterward_event_times are all false. TODO: consider change the code.
                afterward_event_idx = (
                    train_event_times[train_event_indicators == 1] > event_times[i]
                )
                best_guesses[i] = np.mean(
                    train_event_times[train_event_indicators == 1][afterward_event_idx]
                )
        # NaN values are generated because there are no events after the censor times
        nan_idx = np.argwhere(np.isnan(best_guesses))
        predicted_times = np.delete(predicted_times, nan_idx)
        best_guesses = np.delete(best_guesses, nan_idx)
        weights = np.delete(weights, nan_idx)

        if truncated_time:
            best_guesses = np.clip(best_guesses, a_max=truncated_time, a_min=None)
            predicted_times = np.clip(predicted_times, a_max=truncated_time, a_min=None)

        if log_scale:
            errors = np.log(best_guesses) - np.log(predicted_times)
        else:
            errors = best_guesses - predicted_times
        return np.average(error_func(errors), weights=weights)
    elif method == "IPCW-D":
        # This is the IPCW-D method from https://arxiv.org/pdf/2306.01196.pdf
        # Using IPCW weights to transfer the censored subjects to uncensored subjects
        inverse_train_event_indicators = 1 - train_event_indicators

        ipc_model = KaplanMeierArea(train_event_times, inverse_train_event_indicators)
        ipc_pred = ipc_model.predict(event_times)
        # Catch if denominator is 0. This happens when the time is later than the last event time in trainset.
        ipc_pred[ipc_pred == 0] = np.inf

        if truncated_time:
            event_times = np.clip(event_times, a_max=truncated_time, a_min=None)
            predicted_times = np.clip(predicted_times, a_max=truncated_time, a_min=None)

        if log_scale:
            errors = np.log(event_times) - np.log(predicted_times)
        else:
            errors = event_times - predicted_times
        return (
            error_func(errors)[event_indicators] / ipc_pred[event_indicators]
        ).mean()
    elif method == "Pseudo_obs":
        # Calculate the best guess time (surrogate time) by the contribution of the censored subjects to KM curve
        n_train = train_event_times.size

        events, population_counts = (
            km_model.events.copy(),
            km_model.population_count.copy(),
        )
        times = km_model.survival_times.copy()
        probs = km_model.survival_probabilities.copy()
        # get the discrete time points where the event happens, then calculate the area under those discrete time only
        # this doesn't make any difference for step function, but it does for trapezoid rule.
        unique_idx = np.where(events != 0)[0]
        if unique_idx[-1] != len(events) - 1:
            unique_idx = np.append(unique_idx, len(events) - 1)
        times = times[unique_idx]
        population_counts = population_counts[unique_idx]
        events = events[unique_idx]
        probs = probs[unique_idx]
        sub_expect_time = km_mean(times.copy(), probs.copy())

        # use the idea of dynamic programming to calculate the multiplier of the KM estimator in advances.
        # if we add a new time point to the KM curve, the multiplier before the new time point will be
        # 1 - event_counts / (population_counts + 1), and the multiplier after the new time point will be
        # the same as before.
        multiplier = 1 - events / population_counts
        multiplier_total = 1 - events / (population_counts + 1)
        best_guesses = event_times.copy().astype(float)

        for i in trange(
            n_test, desc="Calculating surrogate times for MAE-PO", disable=not verbose
        ):
            if event_indicators[i] != 1:
                total_multiplier = multiplier.copy()
                insert_index = np.searchsorted(times, event_times[i], side="right")
                total_multiplier[:insert_index] = multiplier_total[:insert_index]
                survival_probabilities = np.cumprod(total_multiplier)
                if insert_index == len(times):
                    times_addition = np.append(times, event_times[i])
                    survival_probabilities_addition = np.append(
                        survival_probabilities, survival_probabilities[-1]
                    )
                    total_expect_time = km_mean(
                        times_addition, survival_probabilities_addition
                    )
                else:
                    total_expect_time = km_mean(times, survival_probabilities)
                best_guesses[i] = (
                    n_train + 1
                ) * total_expect_time - n_train * sub_expect_time

        if truncated_time:
            best_guesses = np.clip(best_guesses, a_max=truncated_time, a_min=None)
            predicted_times = np.clip(predicted_times, a_max=truncated_time, a_min=None)

        if log_scale:
            errors = np.log(best_guesses) - np.log(predicted_times)
        else:
            errors = best_guesses - predicted_times
        return np.average(error_func(errors), weights=weights)
    else:
        raise ValueError(
            "Method must be one of 'Uncensored', 'Hinge', 'Margin', 'IPCW-T', 'IPCW-D' "
            "or 'Pseudo_obs'. Got '{}' instead.".format(method)
        )


def insert_km(
    survival_times: np.ndarray,
    event_count: np.ndarray,
    as_risk_count: np.ndarray,
    new_t: float,
    new_e: int,
) -> tuple[np.ndarray, np.ndarray]:
    """
    Insert a new time point into the Kaplan-Meier curve.

    Parameters
    ----------
    survival_times: np.ndarray, shape = (n_samples, )
        Survival times for KM curve of the testing samples
    event_count: np.ndarray, shape = (n_samples, )
        Event count for KM curve of the testing samples at each time point
    as_risk_count: np.ndarray, shape = (n_samples, )
        At-risk count for KM curve of the testing samples at each time point
    new_t: float
        New time point to be inserted
    new_e: int
        New event count to be inserted

    Returns
    -------
    survival_times: np.ndarray, shape = (n_samples, )
        Survival times for KM curve of the testing samples
    survival_probabilities: np.ndarray, shape = (n_samples, )
        Survival probabilities for KM curve of the testing samples
    """
    # Find the index where new_t should be inserted
    insert_index = np.searchsorted(survival_times, new_t)

    # Check if new_t is already at the found index
    if insert_index < len(survival_times) and survival_times[insert_index] == new_t:
        # If new_t is already in the array, increment the event count
        event_count[insert_index] += new_e
        as_risk_count[: insert_index + 1] += 1
    else:
        # Insert new_t into
        survival_times = np.insert(survival_times, insert_index, new_t)
        event_count = np.insert(event_count, insert_index, new_e)
        # if beyond the last time point, manually insert the last at-risk count with 0, then add 1 to all before
        if insert_index == survival_times.size - 1:
            as_risk_count = np.insert(as_risk_count, insert_index, 0)
        else:
            as_risk_count = np.insert(
                as_risk_count, insert_index, as_risk_count[insert_index]
            )
        as_risk_count[: insert_index + 1] += 1

    event_ratios = 1 - event_count / as_risk_count
    survival_probabilities = np.cumprod(event_ratios)

    return survival_times, survival_probabilities


def _prepare_interval_arrays(
    left_bounds: np.ndarray,
    right_bounds: np.ndarray,
    predicted_times: np.ndarray,
) -> tuple[np.ndarray, np.ndarray, np.ndarray]:
    L = np.asarray(left_bounds, float)
    R = np.asarray(right_bounds, float)
    t_hat = np.asarray(predicted_times, float)

    assert L.shape == R.shape == t_hat.shape, "shape mismatch"
    return L, R, t_hat


def _compute_inside_mask(
    left: np.ndarray,
    right: np.ndarray,
    predicted: np.ndarray,
) -> tuple[np.ndarray, np.ndarray]:
    """
    Compute mask for predictions inside their intervals. Also return mask for right-censored intervals.
    Parameters

    Inside test:
    - general finite-interval: (L, R]  -> (t_hat > L) & (t_hat <= R)
    - right-censored:          (L, +inf) -> t_hat > L
    - exact-interval (L==R): treat as {R}: |t_hat - R| <= atol
    ----------
    left: np.ndarray, (n_samples,)
        Left limits of the interval-censored data.
    right: np.ndarray, (n_samples,)
        Right limits of the interval-censored data (use np.inf for right-censor).
    predicted: np.ndarray, (n_samples,)
        Your point predictions (e.g., median predicted times, mean predicted times, etc).

    Returns
    -------
    inside: np.ndarray, (n_samples,)
        Boolean mask for predictions inside their intervals.
    is_right_cens: np.ndarray, (n_samples,)
        Boolean mask for right-censored intervals.
    """
    is_right_cens = np.isinf(right)
    inside_finite = (~is_right_cens) & (predicted > left) & (predicted <= right)
    inside_right = is_right_cens & (predicted > left)
    inside = inside_finite | inside_right
    return inside, is_right_cens


def inclusion_rate(
    left_bounds: np.ndarray,
    right_bounds: np.ndarray,
    predicted_times: np.ndarray,
) -> float:
    """
    Inclusion rate for interval-censored data: proportion of predictions inside their intervals.

    This is similar to p_out but reports the complement (1 - p_out).
    p_out is proposed in [1].

    Parameters
    ----------
    left_bounds: np.ndarray, (n_samples,)
        Left limits of the interval-censored data.
    right_bounds: np.ndarray, (n_samples,)
        Right limits of the interval-censored data (use np.inf for right-censor).
    predicted_times: np.ndarray, (n_samples,)
        Your point predictions (e.g., median predicted times, mean predicted times, etc).

    Returns
    -------
    inclusion_rate: float
        Proportion of samples with predicted times inside their intervals.
    References
    ----------
    [1] Avati et al., "Countdown regression: sharp and calibrated survival predictions", UAI 2020.
    """
    L, R, t_hat = _prepare_interval_arrays(left_bounds, right_bounds, predicted_times)
    inside, _ = _compute_inside_mask(L, R, t_hat)
    return float(np.mean(inside))


def mean_error_ic(
    left_bounds: np.ndarray,
    right_bounds: np.ndarray,
    predicted_times: np.ndarray,
    error_type: str = "absolute",
    log_scale: bool = False,
) -> float:
    """
    Mean error for interval-censored data. It is a one-sided mean error that only penalizes predictions
    outside the interval-censored data.
    This is proposed in [1] as the hinge loss for interval-censored data.
    This is later reporposed by [2] as the `d_out` metric.

    Parameters
    ----------
    left_bounds: np.ndarray, (n_samples,)
        Left limits of the interval-censored data.
    right_bounds: np.ndarray, (n_samples,)
        Right limits of the interval-censored data (use np.inf for right-censor).
    predicted_times: np.ndarray, (n_samples,)
        Your point predictions (e.g., median predicted times, mean predicted times, etc).
    error_type: string, default: "absolute"
        Type of mean error to use. Options are "absolute" and "squared".
    log_scale: boolean, default: False
        Whether to use log scale for the loss function.

    Returns
    -------
    mean_error: float
        Mean error value.

    References
    ----------
    [1] Shivaswamy et al., "A support vector approach to censored targets", ICDM 2007.
    [2] Avati et al., "Countdown regression: sharp and calibrated survival predictions", UAI 2020.

    """
    L, R, t_hat = _prepare_interval_arrays(left_bounds, right_bounds, predicted_times)

    # set the error function
    if error_type == "absolute":
        error_func = np.abs
    elif error_type == "squared":
        error_func = np.square
    else:
        raise TypeError(
            "Please enter one of 'absolute' or 'squared' for calculating error."
        )

    # TODO: We need to move the logarithm transformation later after we calculate a best-guess.
    if log_scale:
        t_hat = np.log(t_hat)
        L = np.log(L)
        R = np.log(R)

    inside, is_right_cens = _compute_inside_mask(L, R, t_hat)
    outside = ~inside
    error_to_L = error_func(t_hat - L)
    error_to_R = np.where(is_right_cens, np.inf, error_func(t_hat - R))
    d_i = np.where(outside, np.minimum(error_to_L, error_to_R), 0.0)

    return float(np.mean(d_i))


if __name__ == "__main__":
    # Test the functions
    train_t = np.array(
        [
            0,
            1,
            2,
            3,
            4,
            5,
            10,
            11,
            12,
            13,
            14,
            15,
            16,
            17,
            18,
            19,
            20,
            21,
            22,
            26,
            27,
            28,
            29,
            30,
            31,
            32,
            33,
            34,
            60,
            61,
            62,
            63,
            64,
            65,
            66,
            67,
            74,
            75,
            79,
            80,
            81,
            82,
            83,
            84,
            85,
            86,
            87,
            88,
            89,
            90,
            91,
            92,
            93,
            98,
            99,
            100,
            101,
            102,
            103,
            104,
            105,
            106,
            107,
            108,
            109,
            110,
            111,
            117,
            118,
            119,
            120,
            120,
            120,
            121,
            121,
            124,
            125,
            126,
            127,
            128,
            129,
            136,
            137,
            138,
            139,
            140,
            141,
            142,
            143,
            144,
            145,
            146,
            147,
            148,
            149,
            155,
            156,
            157,
            158,
            159,
            161,
            182,
            183,
            186,
            190,
            191,
            192,
            192,
            192,
            193,
            194,
            195,
            196,
            197,
            198,
            199,
            200,
            201,
            202,
            203,
            204,
            202,
            203,
            204,
            202,
            203,
            204,
            212,
            213,
            214,
            215,
            216,
            217,
            222,
            223,
            224,
        ]
    )
    train_e = np.array(
        [
            1,
            1,
            0,
            0,
            0,
            0,
            1,
            1,
            0,
            0,
            0,
            0,
            0,
            0,
            1,
            1,
            0,
            0,
            0,
            0,
            0,
            1,
            1,
            1,
            0,
            1,
            1,
            0,
            0,
            1,
            1,
            1,
            1,
            1,
            1,
            1,
            1,
            1,
            1,
            1,
            1,
            1,
            0,
            0,
            1,
            0,
            0,
            0,
            1,
            1,
            0,
            0,
            0,
            0,
            0,
            0,
            1,
            1,
            0,
            0,
            0,
            0,
            0,
            0,
            1,
            0,
            1,
            0,
            1,
            0,
            1,
            0,
            1,
            0,
            1,
            0,
            1,
            0,
            1,
            0,
            1,
            0,
            1,
            0,
            0,
            1,
            1,
            0,
            1,
            1,
            0,
            1,
            1,
            0,
            0,
            0,
            0,
            1,
            1,
            0,
            0,
            0,
            0,
            0,
            0,
            1,
            1,
            0,
            0,
            0,
            0,
            0,
            1,
            1,
            0,
            0,
            0,
            0,
            0,
            0,
            1,
            1,
            0,
            0,
            0,
            0,
            0,
            0,
            1,
            1,
            0,
            0,
            0,
            0,
            0,
            0,
        ]
    )
    t = np.array(
        [
            5,
            10,
            19,
            31,
            43,
            59,
            63,
            75,
            97,
            113,
            134,
            151,
            163,
            176,
            182,
            195,
            200,
            210,
            220,
        ]
    )
    e = np.array([1, 1, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0])
    predict_time = np.array(
        [
            18,
            19,
            5,
            12,
            75,
            100,
            120,
            85,
            36,
            95,
            170,
            41,
            200,
            210,
            260,
            86,
            100,
            120,
            140,
        ]
    )
    # "Margin", "IPCW-T", "IPCW-D", "Pseudo_obs"
    mae_score = mean_error(
        predict_time,
        t,
        e,
        train_t,
        train_e,
        method="Pseudo_obs",
        verbose=True,
        truncated_time=100,
    )
    print(mae_score)
